"""Test backend module

This module exists just for test purposes
"""
import uuid

import xarray as xr  # noqa: F401

from demessaging import main
from demessaging.types.xarray import DataArray

__all__ = ["compute_sum"]


topic = "test_topic_" + uuid.uuid4().urn[9:]


def compute_sum(da: DataArray) -> DataArray:
    """Compute the sum over a data array.

    Parameters
    ----------
    da : DataArray
        The input data array

    Returns
    -------
    DataArray
        The sum of the data array
    """
    return da.sum()


if __name__ == "__main__":
    main(messaging_config=dict(topic=topic))



"""Main module to start a backend module from the command-line."""
from demessaging.backend import BackendModule, main  # noqa: F401
from demessaging.config import configure, registry

from ._version import get_versions

__all__ = ["main", "configure", "registry", "BackendModule"]


__version__ = get_versions()["version"]
del get_versions



"""Test module for the :mod:`demessaging.backend` module."""
import importlib
import inspect
import pathlib
import subprocess as spr
import sys
from textwrap import dedent
from typing import Dict, List, Type

import pytest

from demessaging import backend

try:
    import xarray
except ImportError:
    xarray = None  # type: ignore
else:
    try:
        import netCDF4
    except ImportError:
        netCDF4 = None  # type: ignore


class TestModuleModel:
    """Test class for the :class:`demessaging.backend.BackendModule`."""

    def test_load_all(self) -> None:
        """Test loading everything with __all__."""
        Model = backend.BackendModule.create_model(
            "_test_module", messaging_config=dict(topic="test")
        )
        schema = Model.schema()

        assert "FuncFuncBasic" in schema["definitions"]
        assert "FuncFuncArbitraryType" in schema["definitions"]
        assert "FuncFuncArbitraryModel" in schema["definitions"]

        assert "ClassClass" in schema["definitions"]
        assert "MethClassClassAdd2a" in schema["definitions"]

        assert "_private_func" not in schema["definitions"]
        assert "_PrivateFunc" not in schema["definitions"]
        assert "PrivateFunc" not in schema["definitions"]

    def test_load_function_name(self) -> None:
        """Test loading everything with __all__."""
        Model = backend.BackendModule.create_model(
            "_test_module",
            messaging_config=dict(topic="test"),
            members=["func_basic"],
        )
        schema = Model.schema()

        assert "FuncFuncBasic" in schema["definitions"]
        assert "FuncFuncArbitraryType" not in schema["definitions"]
        assert "ClassClass" not in schema["definitions"]

    def test_load_function(self) -> None:
        """Test loading everything with __all__."""
        from _test_module import func_basic

        Model = backend.BackendModule.create_model(
            "_test_module",
            messaging_config=dict(topic="test"),
            members=[func_basic],
        )
        schema = Model.schema()

        assert "FuncFuncBasic" in schema["definitions"]
        assert "FuncFuncArbitraryType" not in schema["definitions"]
        assert "ClassClass" not in schema["definitions"]

    def test_load_function_model(self) -> None:
        """Test loading everything with __all__."""
        from _test_module import func_basic

        FuncModel = backend.BackendFunction.create_model(func_basic)
        Model = backend.BackendModule.create_model(
            "_test_module",
            messaging_config=dict(topic="test"),
            members=[FuncModel],
        )
        schema = Model.schema()

        assert "FuncFuncBasic" in schema["definitions"]
        assert "FuncFuncArbitraryType" not in schema["definitions"]
        assert "ClassClass" not in schema["definitions"]

    def test_parse_basic(self) -> None:
        """Test parsing a request to the :func:`func_basic` function."""
        Model = backend.BackendModule.create_model(
            "_test_module", messaging_config=dict(topic="test")
        )

        obj = {"func_name": "func_basic", "a": 1}

        model = Model.parse_obj(obj)
        val = model()
        assert val.__root__ == [1]  # type: ignore

    def test_parse_class(self) -> None:
        """Test parsing a request for a class method."""
        Model = backend.BackendModule.create_model(
            "_test_module", messaging_config=dict(topic="test")
        )

        obj = {
            "class_name": "Class",
            "a": 2,
            "function": {"func_name": "add2a", "c": 2},
        }

        model = Model.parse_obj(obj)
        val = model()
        assert val.__root__ == 4  # type: ignore

    def test_connect(
        self, test_module_path: str, test_module_args: List[str]
    ) -> None:
        """Test connecting to the messaging server."""
        spr.check_call(
            [sys.executable, test_module_path]
            + test_module_args
            + ["test-connect"]
        )

    @pytest.mark.usefixtures("connected_module")
    def test_request(
        self,
        test_module_command: List[str],
        test_request_path: str,
    ) -> None:
        """Test parsing a request via the pulsar messaging system."""
        spr.check_call(
            test_module_command + ["send-request", test_request_path]
        )

    @pytest.mark.usefixtures("connected_module")
    def test_request_arbitrary(
        self,
        test_module_command: List[str],
        test_request_arbitrary_path: str,
    ) -> None:
        """Test parsing a request via the pulsar messaging system."""
        spr.check_call(
            test_module_command + ["send-request", test_request_arbitrary_path]
        )

    @pytest.mark.skipif(
        xarray is None or netCDF4 is None,
        reason="xarray and netCDF4 are required",
    )
    @pytest.mark.usefixtures("connected_module_xr")
    def test_request_xarray(
        self,
        test_module_command_xr: List[str],
        test_request_xr_path: str,
    ) -> None:
        """Test parsing a request via the pulsar messaging system."""
        spr.check_call(
            test_module_command_xr + ["send-request", test_request_xr_path]
        )

    @pytest.mark.usefixtures("connected_module_report")
    def test_request_report(
        self,
        test_module_command_report: List[str],
        test_request_report_path: str,
        spr_report_env: Dict,
    ) -> None:
        """Test parsing a request via the pulsar messaging system."""
        spr.check_call(
            test_module_command_report
            + ["send-request", test_request_report_path],
            env=spr_report_env,
        )

    def test_render(
        self, default_class, tmp_module: pathlib.Path, random_mod_name: str
    ) -> None:
        """Test the rendering of a module."""
        import _test_module as ref

        Model = backend.BackendModule.create_model(
            "_test_module", messaging_config=dict(topic="test")
        )

        code = Model.backend_config.render()

        tmp_module.write_text(code)

        mod = importlib.import_module(random_mod_name)

        # test the Class member
        assert hasattr(mod, "Class")

        Class: Type[ref.Class] = mod.Class  # type: ignore

        ref_doc = dedent(inspect.getdoc(ref.Class)).strip()  # type: ignore
        func_doc = dedent(inspect.getdoc(Class)).strip()  # type: ignore

        assert func_doc == ref_doc

        init_ref_doc = dedent(inspect.getdoc(ref.Class.__init__)).strip()  # type: ignore  # noqa: E501
        init_doc = dedent(inspect.getdoc(Class.__init__)).strip()  # type: ignore  # noqa: E501

        assert init_ref_doc == init_doc

        assert hasattr(Class, "sum") and callable(Class.sum)

        params_ref = list(inspect.signature(ref.Class.sum).parameters)
        params = list(inspect.signature(Class.sum).parameters)

        assert params == params_ref

        # test the function member
        assert hasattr(mod, "func_basic")

        ref_doc = dedent(inspect.getdoc(ref.func_basic)).strip()  # type: ignore  # noqa: E501
        func_doc = dedent(inspect.getdoc(mod.func_basic)).strip()  # type: ignore  # noqa: E501

        assert func_doc == ref_doc

    def test_render_with_type(
        self,
        tmp_module: pathlib.Path,
        random_mod_name: str,
    ) -> None:
        """Test the rendering of a module with a custom type defined in it."""

        Model = backend.BackendModule.create_model(
            "_test_module_with_type", messaging_config=dict(topic="test")
        )

        code = Model.generate()

        tmp_module.write_text(code)

        mod = importlib.import_module(random_mod_name)

        # test the Class member
        assert hasattr(mod, "MyType")

        ini = mod.MyType(a=1)  # type: ignore
        assert ini.add2a(1) == 2

    @pytest.mark.usefixtures("connected_module_basemodel")
    def test_generate_and_call_basemodel(
        self,
        test_module_command_basemodel: List[str],
        tmp_module: pathlib.Path,
        random_mod_name: str,
        spr_report_env: Dict,
    ) -> None:
        """Test the generation of a frontend API."""

        with tmp_module.open("w") as f:
            spr.check_call(
                test_module_command_basemodel + ["generate"],
                stdout=f,
                env=spr_report_env,
            )

        mod = importlib.import_module(random_mod_name)

        # test the Class member
        assert hasattr(mod, "TestBaseModelType")

        result = mod.func_basic([dict(data=2)])  # type: ignore
        assert result == 2

        result = mod.TestClass([dict(data=2)]).test_func(dict(data=2))  # type: ignore
        assert result == 4

    @pytest.mark.usefixtures("connected_module")
    def test_generate_and_call(
        self,
        test_module_command: List[str],
        tmp_module: pathlib.Path,
        random_mod_name: str,
    ) -> None:
        """Test the generation of a frontend API."""
        import _test_module as ref

        with tmp_module.open("w") as f:
            spr.check_call(test_module_command + ["generate"], stdout=f)

        mod = importlib.import_module(random_mod_name)

        # test the function member
        assert hasattr(mod, "func_basic")

        result = mod.func_basic(1)  # type: ignore
        assert result == ref.func_basic(1)

        # test the Class member
        assert hasattr(mod, "Class")

        Class: Type[ref.Class] = mod.Class  # type: ignore

        ini = Class(1)
        ref_ini = ref.Class(1)
        result = ini.add2a(2)
        assert result == ref_ini.add2a(2)

    @pytest.mark.usefixtures("connected_module_report")
    def test_generate_and_call_report(
        self,
        test_module_command_report: List[str],
        tmp_module: pathlib.Path,
        random_mod_name: str,
        spr_report_env: Dict,
    ) -> None:
        """Test the generation of a frontend API."""
        import _test_module_with_report as ref

        with tmp_module.open("w") as f:
            spr.check_call(
                test_module_command_report + ["generate"],
                stdout=f,
                env=spr_report_env,
            )

        mod = importlib.import_module(random_mod_name)

        # test the function member
        assert hasattr(mod, "report_test")

        TestReport: Type[ref.TestReport] = mod.TestReport  # type: ignore

        result: TestReport = mod.report_test()  # type: ignore
        assert isinstance(result, TestReport)  # type: ignore

        assert len(TestReport._reports) == 2



"""Custom types for the validation of backend module routines."""
from base64 import b64decode, b64encode
from typing import Union

import xarray as xr

from demessaging.config import registry
from demessaging.types import TypeMixinSlots


class DataArray(xr.DataArray, TypeMixinSlots):
    # an extension of the common :class:`xarray.DataArray` with the necessary
    # classmethods for pydantic

    __slots__ = ()

    __doc__ = xr.DataArray.__doc__

    @classmethod
    def validate(cls, v):
        if isinstance(v, dict):
            return xr.DataArray.from_dict(v)
        elif isinstance(v, str):
            return xr.open_dataarray(b64decode(v.encode("utf-8")))
        elif isinstance(v, bytes):
            return xr.open_dataarray(v)
        else:
            return xr.DataArray(v)


class Dataset(xr.Dataset, TypeMixinSlots):

    __slots__ = ()

    __doc__ = xr.Dataset.__doc__

    @classmethod
    def validate(cls, v):
        if isinstance(v, dict):
            return xr.Dataset.from_dict(v)
        elif isinstance(v, str):
            return xr.open_dataset(b64decode(v.encode("utf-8")))
        elif isinstance(v, bytes):
            return xr.open_dataset(v)
        else:
            return xr.Dataset(v)


def encode_xarray(obj: Union[xr.DataArray, xr.Dataset]) -> bytes:
    netcdf_bytes = obj.to_netcdf(format="NETCDF3_CLASSIC")
    return b64encode(netcdf_bytes).decode("utf-8")  # type: ignore


registry.register_import("demessaging.types.xarray")

for _cls in [DataArray, xr.DataArray, Dataset, xr.Dataset]:
    registry.register_encoder(_cls, encode_xarray)



"""Test module for the :mod:`demessaging.backend` module."""
import importlib
import inspect
import pathlib
from textwrap import dedent

import pytest
from pydantic import ValidationError  # pylint: disable=no-name-in-module

from demessaging import backend

try:
    import xarray
except ImportError:
    xarray = None  # type: ignore
else:
    try:
        import netCDF4
    except ImportError:
        netCDF4 = None  # type: ignore


class TestClassModel:
    """Test functions for creating a class model."""

    def test_class_def(self, default_class) -> None:
        """Test the basic generation of a ClassModel"""
        Model = backend.BackendClass.create_model(default_class)
        config = Model.backend_config

        schema = Model.schema()

        assert schema["title"] == "Class" + default_class.__name__

        assert "a" in schema["properties"]
        assert schema["properties"]["a"]["description"] == "first integer"
        assert "b" in schema["properties"]
        assert schema["properties"]["b"]["description"] == (
            "second integer, by default 1"
        )

        assert set(config.models) == set(["add2a", "add2b", "sum"])
        assert (
            "MethClass" + default_class.__name__ + "Add2a"
            in schema["definitions"]
        )

    def test_private_method(self, default_class) -> None:
        """Test if the private method is excluded."""
        Model = backend.BackendClass.create_model(default_class)
        assert "_private_method" not in Model.backend_config.models

        schema = Model.schema()

        assert "_this_should_not_be_included" not in schema["definitions"]

    def test_valid_method_request(self, default_class) -> None:
        """Test parsing a class request."""
        Model = backend.BackendClass.create_model(default_class)
        obj = {
            "class_name": default_class.__name__,
            "a": 2,
            "function": {"func_name": "add2a", "c": 2},
        }
        Model.parse_obj(obj)

    def test_invalid_method_request(self, default_class) -> None:
        """Test parsing an invalid class request."""
        Model = backend.BackendClass.create_model(default_class)
        obj = {
            "class_name": default_class.__name__,
            "a": 2,
            "function": {"func_name": "asd", "c": 2},
        }
        with pytest.raises(ValidationError):
            Model.parse_obj(obj)

    def test_call_method(self, default_class) -> None:
        """Test parsing a class request."""
        Model = backend.BackendClass.create_model(default_class)
        obj = {
            "class_name": default_class.__name__,
            "a": 2,
            "function": {"func_name": "add2a", "c": 2},
        }
        model = Model.parse_obj(obj)
        ret = model()
        assert ret.__root__ == 4  # type: ignore

    def test_explicit_method_name(self, default_class) -> None:
        """Test parsing one method explicitly."""
        Model = backend.BackendClass.create_model(
            default_class, methods=["add2a"]
        )

        assert list(Model.backend_config.models) == ["add2a"]

    def test_explicit_method(self, default_class) -> None:
        """Test parsing one method explicitly."""
        Model = backend.BackendClass.create_model(
            default_class, methods=[default_class.add2a]
        )

        assert list(Model.backend_config.models) == ["add2a"]

    def test_explicit_method_model(self, default_class) -> None:
        """Test parsing one method explicitly."""
        FuncModel = backend.BackendFunction.create_model(default_class.add2a)
        Model = backend.BackendClass.create_model(
            default_class, methods=[FuncModel]
        )

        assert list(Model.backend_config.models) == ["add2a"]

    def test_render(
        self, default_class, tmp_module: pathlib.Path, random_mod_name: str
    ) -> None:
        """Test the rendering of a class."""
        Model = backend.BackendClass.create_model(default_class)

        code = Model.backend_config.render()

        tmp_module.write_text("from __future__ import annotations\n" + code)

        mod = importlib.import_module(random_mod_name)

        assert hasattr(mod, default_class.__name__)
        Class = getattr(mod, default_class.__name__)

        ref_doc = dedent(inspect.getdoc(default_class)).strip()  # type: ignore
        func_doc = dedent(inspect.getdoc(Class)).strip()  # type: ignore

        assert func_doc == ref_doc

        init_ref_doc = dedent(inspect.getdoc(default_class.__init__)).strip()  # type: ignore  # noqa: E501
        init_doc = dedent(inspect.getdoc(Class.__init__)).strip()  # type: ignore  # noqa: E501

        assert init_ref_doc == init_doc

        assert hasattr(Class, "sum") and callable(Class.sum)

        params_ref = list(inspect.signature(default_class.sum).parameters)
        params = list(inspect.signature(Class.sum).parameters)

        assert params == params_ref



import os
import pathlib
import random
import subprocess as spr
import sys
import time
import uuid
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterator, List, Type

import pytest
from pydantic import BaseModel  # pylint: disable=no-name-in-module

from demessaging import config
from demessaging.types import TypeMixin

try:
    import xarray
except ImportError:
    xarray = None  # type: ignore
else:
    try:
        import netCDF4
    except ImportError:
        netCDF4 = None  # type: ignore


os.environ["DASF_REPORT_SHOW_METHOD"] = "PRINT"


if TYPE_CHECKING:
    from _pytest.monkeypatch import MonkeyPatch


test_dir = pathlib.Path(__file__).parent


@pytest.fixture(scope="session")
def test_module_path() -> str:
    return str(test_dir / "_test_module.py")


@pytest.fixture(scope="session")
def test_module_path_xr() -> str:
    return str(test_dir / "_test_module_with_xarray.py")


@pytest.fixture(scope="session")
def test_module_path_report() -> str:
    return str(test_dir / "_test_module_with_report.py")


@pytest.fixture(scope="session")
def test_module_path_basemodel() -> str:
    return str(test_dir / "_test_module_with_basemodel_type.py")


@pytest.fixture(scope="session")
def requests_path() -> pathlib.Path:
    return test_dir / "requests"


@pytest.fixture(scope="session")
def test_request_path(requests_path: pathlib.Path) -> str:
    return str(requests_path / "test_request.json")


@pytest.fixture(scope="session")
def test_request_arbitrary_path(requests_path: pathlib.Path) -> str:
    return str(requests_path / "test_request_arbitrary.json")


@pytest.fixture(scope="session")
def test_request_xr_path(requests_path: pathlib.Path) -> str:
    return str(requests_path / "test_request_xarray.json")


@pytest.fixture(scope="session")
def test_request_report_path(requests_path: pathlib.Path) -> str:
    return str(requests_path / "test_report_request.json")


@pytest.fixture
def tmp_sys_path(
    tmp_path: pathlib.Path, monkeypatch: "MonkeyPatch"
) -> pathlib.Path:
    """Get a temporary path and add it to the sys.path."""
    monkeypatch.syspath_prepend(str(tmp_path))
    return tmp_path


@pytest.fixture
def random_mod_name() -> str:
    """Generate a random module name."""
    return "test_" + str(random.randint(1, 10000))


@pytest.fixture
def tmp_module(tmp_sys_path: pathlib.Path, random_mod_name: str):
    """Generate a path to a module file for tests."""
    return tmp_sys_path / (random_mod_name + ".py")


@pytest.fixture
def func_sig() -> Callable[[int], List[int]]:
    """Create a test function with signature and docstring."""

    def func(a: int) -> List[int]:
        """A test function.

        With some docu.

        Parameters
        ----------
        a: integer
            An integer

        Returns
        -------
        list of int
            A list of integers
        """
        return [a]

    return func


@pytest.fixture
def func_invalid_sig() -> Callable[[int], List[int]]:
    """Create a test function with signature and docstring."""

    def func(a: int) -> List[int]:
        """A test function.

        With some docu.

        Parameters
        ----------
        a: integer
            An integer

        Returns
        -------
        list of int
            A list of integers
        """
        return a  # type: ignore

    return func


@pytest.fixture
def func_missing_doc() -> Callable[[int], int]:
    """Get a function with missing documentation."""

    def func(a: int) -> int:
        return a

    return func


@pytest.fixture
def func_missing_sig() -> Callable:
    """Get a function with missing signatures"""

    def func(a):
        """Do something.

        Parameters
        ----------
        a: int
            A parameter

        Returns
        -------
        int
            An integer
        """
        return a

    return func


class ArbitraryType(TypeMixin):
    def __init__(self, a: int):
        self.a = a

    @classmethod
    def validate(cls, v):
        if isinstance(v, cls):
            return v
        else:
            return cls(int(v))


class ArbitraryTestModel(BaseModel):
    """A model for test purposes."""

    int_test_required: int

    int_test_optional: int = 1

    custom_type_required: ArbitraryType

    # the following does not work!!!
    # because Object of type 'ArbitraryType' is not JSON serializable
    # custom_type_optional: ArbitraryType = ArbitraryType(2)


def encode_arbitrary_type(a: ArbitraryType) -> int:
    return a.a


config.registry.register_encoder(ArbitraryType, encode_arbitrary_type)
config.registry.register_import("conftest")


default_registry = config.registry.copy()


@pytest.fixture(autouse=True)
def clean_registry() -> Iterator[config.ApiRegistry]:
    """Generate a clean registry for the test."""
    from demessaging import config

    saved_registry = config.registry

    config.registry = default_registry.copy(deep=True)

    yield config.registry

    config.registry = saved_registry


@pytest.fixture
def func_arbitrary_types() -> Callable[[ArbitraryType], ArbitraryType]:
    """Define a function with an arbitraty class as annotation."""

    def func(a: ArbitraryType) -> ArbitraryType:
        return a

    return func


@pytest.fixture
def func_arbitrary_model() -> Callable[
    [ArbitraryTestModel], ArbitraryTestModel
]:
    """Define a function with an arbitrary model."""

    def func(a: ArbitraryTestModel) -> ArbitraryTestModel:
        return a

    return func


@pytest.fixture
def valid_obj() -> Dict[str, Any]:
    return {"a": 1}


@pytest.fixture
def invalid_obj() -> Dict[str, Any]:
    return {"a": "x"}


@pytest.fixture
def valid_arbitrary_model() -> Dict[str, ArbitraryTestModel]:
    return {
        "a": ArbitraryTestModel(
            int_test_required=1,
            custom_type_required=1,  # type: ignore
        )
    }


@pytest.fixture
def default_class() -> Type[object]:
    """Create a test class with two methods."""

    class Class:
        """A test class."""

        def __init__(self, a: int, b: int = 1):
            """
            Parameters
            ----------
            a : int
                first integer
            b : int, optional
                second integer, by default 1
            """
            self.a = a
            self.b = b

        @staticmethod
        def sum(a: int, b: int) -> int:
            """
            Parameters
            ----------
            a : int
                first integer
            b : int, optional
                second integer, by default 1

            Returns
            -------
            int
                The sum of `a` and `b`
            """
            return a + b

        def _this_should_not_be_included(self, a: int) -> int:
            """This method should not be included because it stats with _

            Parameters
            ----------
            a: int
                Some dummy int
            """
            return a

        def add2a(self, c: int) -> int:
            """Add a number to `a`.

            Parameters
            ----------
            c : int
                The number to add

            Returns
            -------
            int
                a + c
            """
            return self.a + c

        def add2b(self, c: int) -> int:
            """Add a number to `b`

            Parameters
            ----------
            c : int
                The number to add

            Returns
            -------
            int
                b + c
            """
            return self.b + c

    return Class


@pytest.fixture(scope="session")
def topic() -> str:
    """Generate a random topic for testing."""
    return "test_topic_" + uuid.uuid4().urn[9:]


@pytest.fixture(scope="session", params=["", "websocket-url"])
def test_module_args(request, topic: str):
    if request.param == "websocket-url":
        return [
            "--websocket-url",
            config.PulsarConfig(topic=topic).get_topic_url("")[:-1],
            "--consumer-url",
            config.PulsarConfig(topic=topic).get_topic_url(
                "", subscription="1"
            )[:-2],
        ]
    else:
        return []


@pytest.fixture(scope="session")
def test_module_command(
    test_module_path: str, topic: str, test_module_args: List[str]
) -> List[str]:
    """Generate a command for subprocess to launch the test module."""
    return [sys.executable, test_module_path, "-t", topic] + test_module_args


@pytest.fixture(scope="session")
def connected_module(test_module_command: List[str]) -> Iterator[spr.Popen]:
    """Generate a connection to the pulsar messaging system."""
    background_process = None
    try:
        background_process = spr.Popen(test_module_command + ["listen"])
    except Exception:
        raise
    else:
        time.sleep(1)
        yield background_process
        background_process.terminate()


@pytest.fixture(scope="session")
def topic_xr() -> str:
    """Generate a random topic for testing."""
    return "test_topic_" + uuid.uuid4().urn[9:]


@pytest.fixture(scope="session")
def test_module_command_xr(
    test_module_path_xr: str, topic_xr: str
) -> List[str]:
    """Generate a command for subprocess to launch the test module."""
    return [sys.executable, test_module_path_xr, "-t", topic_xr]


@pytest.fixture(scope="session")
def topic_basemodel() -> str:
    """Generate a random topic for testing."""
    return "test_topic_" + uuid.uuid4().urn[9:]


@pytest.fixture(scope="session")
def topic_report() -> str:
    """Generate a random topic for testing."""
    return "test_topic_" + uuid.uuid4().urn[9:]


@pytest.fixture(scope="session")
def test_module_command_report(
    test_module_path_report: str, topic_report: str
) -> List[str]:
    """Generate a command for subprocess to launch the test module."""
    return [sys.executable, test_module_path_report, "-t", topic_report]


@pytest.fixture(scope="session")
def test_module_command_basemodel(
    test_module_path_basemodel: str, topic_basemodel: str
) -> List[str]:
    """Generate a command for subprocess to launch the test module."""
    return [sys.executable, test_module_path_basemodel, "-t", topic_basemodel]


@pytest.fixture(scope="session")
def connected_module_xr(
    test_module_command_xr: List[str],
) -> Iterator[spr.Popen]:
    """Generate a connection to the pulsar messaging system."""
    background_process = None

    try:
        background_process = spr.Popen(test_module_command_xr + ["listen"])
    except Exception:
        raise
    else:
        yield background_process
        background_process.terminate()


@pytest.fixture(scope="session")
def spr_basemodel_env(topic_basemodel: str) -> Dict:
    spr_env = os.environ.copy()
    spr_env["TEST_TOPIC"] = topic_basemodel
    return spr_env


@pytest.fixture(scope="session")
def connected_module_basemodel(
    test_module_command_basemodel: List[str], spr_basemodel_env: Dict
) -> Iterator[spr.Popen]:
    """Generate a connection to the pulsar messaging system."""
    background_process = None

    try:
        background_process = spr.Popen(
            test_module_command_basemodel + ["listen"], env=spr_basemodel_env
        )
    except Exception:
        raise
    else:
        yield background_process
        background_process.terminate()


@pytest.fixture(scope="session")
def spr_report_env(topic_report: str) -> Dict:
    spr_env = os.environ.copy()
    spr_env["TEST_TOPIC"] = topic_report
    return spr_env


@pytest.fixture(scope="session")
def connected_module_report(
    test_module_command_report: List[str], spr_report_env: Dict
) -> Iterator[spr.Popen]:
    """Generate a connection to the pulsar messaging system."""
    background_process = None

    try:
        background_process = spr.Popen(
            test_module_command_report + ["listen"], env=spr_report_env
        )
    except Exception:
        raise
    else:
        yield background_process
        background_process.terminate()



"""Test backend module

This module exists just for test purposes
"""
import os
import uuid
from typing import ClassVar, List, Literal, Optional, cast

from deprogressapi import PrintReport

from demessaging import config, main

__all__ = ["report_test"]


topic = os.getenv("TEST_TOPIC", "test_topic_" + uuid.uuid4().urn[9:])


config.registry.hard_code(f"topic = '{topic}'")


@config.registry.register_type
class TestReport(PrintReport):
    """Just a test reporter class to see the messaging works."""

    _reports: ClassVar[List[PrintReport]] = []

    report_type: Literal["_test_report_" + topic] = "_test_report_" + topic  # type: ignore

    def show_print(self):
        self._reports.append(self)


def report_test(
    reporter: Optional[TestReport] = TestReport(report_id="test_report"),
) -> TestReport:
    """Test function for using a report via the pulsar.

    Parameters
    ----------
    testreport : Optional[TestReport]
        A reporter object for the test
    """
    reporter = cast(TestReport, TestReport.from_arg(reporter))
    reporter.print("Test report")
    reporter.error("Error report")
    reporter.complete()
    return reporter


if __name__ == "__main__":
    main(messaging_config=dict(topic=topic))



"""Test module for :mod:`demessaging.config`."""
from typing import Type

from demessaging import backend, config


def test_configure_function(func_sig) -> None:
    """Test configurating a function."""
    config.configure(field_params={"a": {"gt": 0}})(func_sig)
    Model = backend.BackendFunction.create_model(func_sig)

    schema = Model.schema()

    assert "exclusiveMinimum" in schema["properties"]["a"]
    assert schema["properties"]["a"]["exclusiveMinimum"] == 0


def test_configure_class(default_class) -> None:
    """Test configurating a function."""
    config.configure(field_params={"a": {"gt": 0}})(default_class)
    Model = backend.BackendClass.create_model(default_class)

    schema = Model.schema()

    assert "exclusiveMinimum" in schema["properties"]["a"]
    assert schema["properties"]["a"]["exclusiveMinimum"] == 0


def test_configure_method(default_class: Type[object]) -> None:
    """Test configuring a specific method."""

    class TestClass(default_class):  # type: ignore
        @config.configure(field_params={"a": {"gt": 0}})
        def test_method(self, a: int) -> int:
            return a

    Model = backend.BackendClass.create_model(TestClass)

    schema = Model.schema()

    assert (
        "exclusiveMinimum"
        in schema["definitions"]["MethClassTestClassTestMethod"]["properties"][
            "a"
        ]
    )
    aconf = schema["definitions"]["MethClassTestClassTestMethod"][
        "properties"
    ]["a"]
    assert aconf["exclusiveMinimum"] == 0



"""Example backend module for the de-messaging framework.

Start this module by typing::

    python ExampleMessageConsumer.py -t your-custom-topic connect

in the terminal, or::

    python ExampleMessageConsumer.py --help

for further options.
"""
import datetime
from typing import List, Optional

# this import is (unfortunately) necessary to resolve the deprogressapi.tree in
# the generated client stub
import deprogressapi.tree  # noqa: F401
from deprogressapi import ProgressReport
from pydantic import BaseModel, validate_arguments

from demessaging import configure, main, registry
from demessaging.types.xarray import DataArray

__all__ = ["hello_world", "HelloWorld", "compute_sum", "report_progress_demo"]


@registry.register_type
class GreetResponse(BaseModel):

    message: str
    greetings: List[str]
    greeting_time: datetime.datetime


# You can configure different validations for the parameters.
# Here, we know that `repeat` must at least be 0, so we set this as a field
# parameter (see
# https://pydantic-docs.helpmanual.io/usage/schema/#field-customisation)
# for available validators
@configure(field_params={"repeat": {"ge": 0}})
def hello_world(
    message: str, repeat: int, greet_message: str
) -> GreetResponse:
    """Greet the hello world module.

    Parameters
    ----------
    message: str
        Message that will be part of the greet response
    repeat: int
        Number of repetitions of the greet message
    greet_message: str
        Message string that will be mirror back to the requesting node

    Returns
    -------
    GreetResponse
        Hello world greet response object
    """
    greetings: List[str] = [greet_message] * repeat
    return GreetResponse(
        message=message,
        greetings=greetings,
        greeting_time=datetime.datetime.now(),
    )


# You can also use generic python classes as input/output for your model. In
# this case, we use xarrays DataArray. Note that you then have to define a
# special `validate` method. See :class:`demessaging.types.xarray.DataArray`
# for instance.
@validate_arguments
def compute_sum(
    da: DataArray,
    reporter: ProgressReport = ProgressReport(step_message="Sum computation"),
) -> DataArray:
    """Compute the sum over a data array.

    Parameters
    ----------
    da : DataArray
        The input data array

    Returns
    -------
    DataArray
        The sum of the data array
    """
    import time

    with reporter:
        for i in range(10):
            with reporter.create_subreport(step_message=f"Report {i}"):
                time.sleep(1)
        ret = da.sum()
    return ret


def report_progress_demo(
    reporter: Optional[ProgressReport] = ProgressReport(
        step_message="Demo report"
    ),
) -> None:
    """Demo function for a progress report with the dasf-progress-api."""
    import time

    reporter = ProgressReport.from_arg(reporter)

    num_steps = 10
    reporter.steps = num_steps

    with reporter:
        for i in range(num_steps):
            with reporter.create_subreport(step_message=f"Sub report {i}"):
                time.sleep(1)
    return


# You can configure how the individual class in the backend module shall be
# interpreted with the `configure` function. See
# :class:`demessaging.config.ClassConfig` for available parameters
@configure(methods=["repeat_message"])
class HelloWorld:
    """Greet the world from a class.

    Classes can define the methods that shall be used for the backend, and they
    define a constructor."""

    def __init__(self, message: str):
        """
        Parameters
        ----------
        message: str
            The hello message
        """
        self.message = message

    @configure(field_params={"repeat": {"ge": 0}})
    def repeat_message(self, repeat: int) -> List[str]:
        """Repeat the `message` multiple times.

        Parameters
        ----------
        repeat: int
            The number of times, `message` shall be repeated.
        """
        return [self.message] * repeat


if __name__ == "__main__":
    main(
        messaging_config=dict(topic="mytesttopic", max_workers=4, queue_size=4)
    )



"""Utitlity functions for the backend framework."""
from __future__ import annotations

import asyncio
import inspect
import threading
import warnings
from itertools import chain, starmap
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Tuple,
    Type,
    Union,
    get_args,
    get_origin,
)

from deprogressapi import BaseReport
from pydantic import Field  # pylint: disable=no-name-in-module

try:
    from typing import Literal  # type: ignore
except ImportError:
    from typing_extensions import Literal  # type: ignore


if TYPE_CHECKING:
    import docstring_parser
    import isort.identify

    from demessaging.config import ClassConfig, FunctionConfig


def get_kws(sig, obj) -> Dict[str, Any]:
    """Get keywords from a signature and a base model."""
    return {
        param: getattr(obj, param)
        for param in sig.parameters.keys()
        if param != "self"
    }


def get_fields(
    name: str,
    sig: inspect.Signature,
    docstring: docstring_parser.Docstring,
    config: Union[FunctionConfig, ClassConfig],
) -> Dict[str, Tuple[Any, Any]]:
    """Get the model fields from a function signature.

    Parameters
    ----------
    name: str
        The name of the function or class
    sig: inspect.Signature
        The signature of the callable
    docstring: docstring_parser.Docstring
        The parser that analyzed the docstring
    config: FunctionConfig or ClassConfig
        The configuration for the callable

    Returns
    -------
    dict
        A mapping from field name to field parameters to be used in
        :func:`pydantic.create_model`.
    """
    fields: Dict[str, Tuple[Any, Any]] = {
        "func_name": (
            Literal[name],  # type: ignore
            Field(description=f"The name of the function. Must be {name!r}"),
        ),
    }
    for key, param in sig.parameters.items():
        if key == "self":
            continue
        field_kws: Dict[str, Any] = {}
        if param.default is not param.empty:
            field_kws["default"] = param.default
        param_doc = next(
            (p for p in docstring.params if p.arg_name == key), None
        )
        if param_doc is not None:
            field_kws["description"] = param_doc.description

        if key in config.field_params:
            field_kws.update(config.field_params[key])

        if key in config.annotations:
            fields[key] = (config.annotations[key], Field(**field_kws))  # type: ignore
        elif param.annotation is param.empty:
            warnings.warn(
                f"Missing signature for {key}, so no validation will "
                "be made for this parameter!",
                RuntimeWarning,
            )
            fields[key] = (Any, Field(**field_kws))  # type: ignore
        else:
            fields[key] = (param.annotation, Field(**field_kws))  # type: ignore

        # test for dasf-progress-api reports and add them to the config
        if param.annotation is not param.empty and _is_progress_report(
            param.annotation
        ):
            if param.default is not param.empty:
                config.reporter_args[key] = param.default
            else:
                config.reporter_args[key] = param.annotation()

    return fields


def _is_progress_report(cls_: Type) -> bool:
    if get_origin(cls_):
        # we do have a Union-type
        return any(
            inspect.isclass(c) and issubclass(c, BaseReport)
            for c in get_args(cls_)
        )
    elif inspect.isclass(cls_):
        return issubclass(cls_, BaseReport)
    else:
        return False


def get_desc(docstring: docstring_parser.Docstring) -> str:
    """Get the description of an object.

    Parameters
    ----------
    docstring: docstring_parser.Docstring
        The parser that analyzed the docstring.

    Returns
    -------
    str
        The description of the callable.
    """
    desc = ""
    if docstring.short_description:
        desc += docstring.short_description
    if docstring.long_description:
        if docstring.blank_after_short_description:
            desc += "\n\n"
        else:
            desc += "\n"
        desc += docstring.long_description
    return desc.strip()


def camelize(w: str) -> str:
    """Camelize a word by making the first letter upper case."""
    return w and (w[:1].upper() + w[1:])


def snake_to_camel(*words: str) -> str:
    """Transform a list of words into its camelized version."""
    return "".join(
        map(camelize, chain.from_iterable(w.split("_") for w in words))
    )


class AsyncIoThread(threading.Thread):
    """A thread that runs an async function.

    See: func: `run_async` for the implementation."""

    def __init__(self, func: Callable, args: Tuple, kwargs: Dict):
        self.__func = func
        self.__args = args
        self.__kwargs = kwargs
        super().__init__()

    def run(self):
        self.result = asyncio.run(self.__func(*self.__args, **self.__kwargs))


def run_async(func: Callable, *args: Any, **kwargs: Any) -> Any:
    """Run an async function and wait for the result.

    This function works within standard python scripts, and during a running
    jupyter session."""
    # check if we have a running loop (which is the case for a jupyter
    # notebook)
    loop: Any
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = None

    if loop and loop.is_running():  # jupyter notebook
        thread = AsyncIoThread(func, args, kwargs)
        thread.start()
        thread.join()  # wait for the thread to finish
        return thread.result
    else:  # standard python script
        return asyncio.run(func(*args, **kwargs))


class ImportMixin:
    """Mixin class for :class:`isort.identify.Import`.

    A response to https://github.com/PyCQA/isort/issues/1641.
    """

    def statement(self: isort.identify.Import) -> str:  # type: ignore
        import_cmd = "cimport" if self.cimport else "import"
        if self.attribute:
            import_string = f"from {self.module} {import_cmd} {self.attribute}"
        else:
            import_string = f"{import_cmd} {self.module}"
        if self.alias:
            import_string += f" as {self.alias}"
        return import_string


def get_module_imports(mod: Any) -> str:
    """Get all the imports from a module

    Parameters
    ----------
    mod: module
        The module to use
    """
    try:
        from isort.api import find_imports_in_code
        from isort.identify import Import as BaseImport
    except (ImportError, ModuleNotFoundError):
        return ""

    code = inspect.getsource(mod)

    class Import(ImportMixin, BaseImport):
        pass

    # We could use the :meth:`~isort.identify.Import.statement` method here,
    # but this would not work always (see
    # https://github.com/PyCQA/isort/issues/1641)
    imports = starmap(Import, find_imports_in_code(code))

    return "\n".join(i.statement() for i in imports)



from warnings import warn

from demessaging.messaging.consumer import MessageConsumer  # noqa: F403, F401

warn(
    "The demessaging.PulsarMessageConsumer module has been renamed to "
    "demessaging.messaging.consumer and will be removed soon!",
    DeprecationWarning,
)


class PulsarMessageConsumer(MessageConsumer):
    # deprecated

    def __init__(self, *args, **kwargs):
        warn(
            "The `demessaging.PulsarMessageConsumer.PulsarMessageConsumer` "
            "class has been replaced by the "
            "`demessaging.messaging.consumer.MessageConsumer` class "
            "and will be removed soon!",
            DeprecationWarning,
        )
        super().__init__(*args, **kwargs)



import sys

from demessaging import main
from demessaging.cli import UNKNOWN_MODULE


def _main():
    sys.path.insert(0, ".")
    return main(UNKNOWN_MODULE)


if __name__ == "__main__":
    _main()



"""Base module for a websocket connection."""
from __future__ import annotations

import random
import string
from abc import ABC
from typing import Optional

import websocket
from pydantic import validate_arguments

from demessaging import config


def get_random_letters(length: int) -> str:
    return "".join(random.choice(string.ascii_letters) for i in range(length))


class WebsocketConnection(ABC):
    """Base class to connect to a message broker using a websocket."""

    @validate_arguments
    def __init__(self, pulsar_config: config.BaseMessagingConfig):
        self.pulsar_config = pulsar_config

    def generate_response_topic(self, topic: Optional[str] = None) -> str:
        topic_name = topic or self.pulsar_config.topic or "anonymous"

        return topic_name + "_" + get_random_letters(8)

    def open_socket(
        self,
        subscription: Optional[str] = None,
        topic: Optional[str] = None,
        **connection_kws,
    ) -> websocket.WebSocket:
        topic_name = topic or self.pulsar_config.topic

        topic_url = self.pulsar_config.get_topic_url(topic_name, subscription)

        sock = websocket.create_connection(topic_url, **connection_kws)

        if sock:
            print("connection to {0} established".format(topic_url))

        return sock



"""Configuration classes for the de-messaging backend module."""
from __future__ import annotations

import importlib
import inspect
import textwrap
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
)
from warnings import warn

from deprogressapi import BaseReport
from pydantic import BaseModel  # pylint: disable=no-name-in-module
from pydantic import (
    BaseSettings,
    Field,
    Json,
    PositiveInt,
    root_validator,
    validator,
)

from demessaging.template import Template

if TYPE_CHECKING:
    from demessaging.backend.class_ import BackendClass
    from demessaging.backend.function import BackendFunction


T = TypeVar("T", bound=Callable[..., Any])


def type_to_string(type_: Any):
    if inspect.isclass(type_):
        if type_.__module__ == "builtins":
            return type_.__name__
        return f"{type_.__module__}.{type_.__name__}"
    else:
        return str(type_)


def build_parameter_docs(model: Type[BaseModel]) -> str:
    """Build the docstring for the parameters of a model."""
    docstring = "\n\nParameters\n----------"
    for fieldname, field in model.__fields__.items():
        param_doc = textwrap.dedent(
            f"""
            {fieldname} : {type_to_string(field.type_)}
                {field.field_info.description}
            """
        )
        docstring = docstring + param_doc.rstrip()
    return docstring


def append_parameter_docs(model: Type[BaseModel]) -> Type[BaseModel]:
    """Append the parameters section to the docstring of a model."""
    docstring = build_parameter_docs(model)
    model.__doc__ += docstring
    return model


@append_parameter_docs
class ApiRegistry(BaseModel):
    """A registry for imports and encoders"""

    @validator("imports")
    @classmethod
    def can_import_import(cls, imports: Dict[str, str]) -> Dict[str, str]:
        errors: List[ImportError] = []
        for key in imports:
            try:
                importlib.import_module(key)
            except ImportError as e:
                errors.append(e)
            except Exception:
                raise
        if errors:
            raise ValueError(
                "Could not import all modules!\n    "
                + "\n    ".join(map(str, errors))
            )
        return imports

    json_encoders: Dict[Any, Callable[[Any], Any]] = Field(
        default_factory=dict,
        description=(
            "Custom encoders for datatypes. See "
            "https://pydantic-docs.helpmanual.io/usage/exporting_models/#json_encoders"  # noqa: E501
        ),
    )

    imports: Dict[str, str] = Field(
        default_factory=dict,
        description=(
            "Modules to import at the top of every file. The first "
            "item is the module, the second is the alias"
        ),
    )

    objects: List[str] = Field(
        default_factory=list,
        description=(
            "Source code for objects that should be inlined in the generated "
            "Python API."
        ),
    )

    def register_import(
        self, module: str, alias: Optional[str] = None
    ) -> None:
        """Register a module that needs to be imported in generated API files.

        Parameters
        ----------
        module: str
            The name of the module, e.g. matplotlib.pyplot
        """
        self.imports[module] = alias or ""

    def register_encoder(
        self, Class: Any, encoder: Callable[[Any], Any]
    ) -> None:
        """Register an encoder for the backend config.

        This function can be used to register a custom encoder for a type.
        You should use this for any object that cannot be decoded by default
        using the standard json.dumps.

        Under the hood, this is then transformed as the ``json_encoders``
        configuration value for pydantic (see
        https://pydantic-docs.helpmanual.io/usage/exporting_models/#modeljson)

        Parameters
        ----------
        Class: object
            The type that can be encoded by the given encoder
        encoder: function
            A function that takes one argument, an instance of `Class` and
            converts it to a JSON-conform representation
        """
        self.json_encoders[Class] = encoder

    def register_type(self, obj: T) -> T:
        """Register a class or function to be available in the generated API.

        Use this function if you want to have certain functions of classes
        available in the generated API module, but they should not be
        transformed to a call to the backend module.
        """
        self.objects.append(inspect.getsource(obj))
        return obj

    def hard_code(self, python_code: str) -> None:
        """Register some code to be implemented in the generated module.

        Parameters
        ----------
        python_code: str
            The code that is supposed to be executed on a module level.
        """
        self.objects.append(python_code)
        return


#: registry for the stuff that should be available in the generated client stub
registry = ApiRegistry()


def _get_registry() -> ApiRegistry:
    """Convenience function to get :attr:`registry`.

    Without this, the default value for the config classes `registry` attribute
    would always be empty.
    """
    return registry


@append_parameter_docs
class BaseConfig(BaseModel):
    """Configuration base class for functions, modules and classes."""

    doc: str = Field(
        "",
        description=(
            "The documentation of the object. If empty, this will be taken "
            "from the corresponding ``__doc__`` attribute."
        ),
    )

    registry: ApiRegistry = Field(
        default_factory=_get_registry,
        description="Utilities for imports and encoders.",
    )

    template: Template = Field(
        Template(name="empty"),
        description=(
            "The :class:`demessaging.template.Template` that is used "
            "to render this object for the generated API."
        ),
    )

    def render(self, **context) -> str:
        """Generate the code to call this function in the frontend."""
        context["config"] = self
        code = self.template.render(**context)
        return code


@append_parameter_docs
class FunctionConfig(BaseConfig):
    """Configuration class for a backend module function."""

    class Config:
        arbitrary_types_allowed = True
        extra = "forbid"

    name: str = Field(
        "",
        description=(
            "The name of the function. If empty, this will be taken from the "
            "functions ``__name__`` attribute."
        ),
    )

    signature: Optional[inspect.Signature] = Field(
        None,
        description=(
            "The calling signature for the function. If empty, this will be "
            "taken from the function itself."
        ),
    )

    validators: Dict[str, classmethod] = Field(
        default_factory=dict,
        description="custom validators for the constructor parameters",
    )

    field_params: Dict[str, Dict[str, Any]] = Field(
        default_factory=dict,
        description=(
            "custom Field overrides for the constructor parameters. See "
            ":func:`pydantic.Fields.Field`"
        ),
    )

    returns: Dict[str, Any] = Field(
        default_factory=dict, description="custom returns overrides."
    )

    annotations: Dict[str, Any] = Field(
        default_factory=dict,
        description="custom annotations for function parameters",
    )

    template: Template = Field(
        Template(name="function.py"),
        description=(
            "The :class:`demessaging.template.Template` that is used "
            "to render the function for the generated API."
        ),
    )

    reporter_args: Dict[str, BaseReport] = Field(
        default_factory=dict,
        description="Arguments that use the dasf-progress-api",
    )


@append_parameter_docs
class ClassConfig(BaseConfig):
    """Configuration class for a backend module class."""

    class Config:
        arbitrary_types_allowed = True
        extra = "forbid"

    name: str = Field(
        "",
        description=(
            "The name of the function. If empty, this will be taken from the "
            "classes ``__name__`` attribute."
        ),
    )

    init_doc: str = Field(
        "",
        description=(
            "The documentation of the function. If empty, this will be taken "
            "from the classes ``__init__`` method."
        ),
    )

    signature: Optional[inspect.Signature] = Field(
        None,
        description=(
            "The calling signature for the function. If empty, this will be "
            "taken from the function itself."
        ),
    )

    methods: List[str] = Field(
        default_factory=list,
        description="methods to use within the backend modules",
    )

    validators: Dict[str, classmethod] = Field(
        default_factory=dict,
        description="custom validators for the constructor parameters",
    )

    field_params: Dict[str, Dict[str, Any]] = Field(
        default_factory=dict,
        description=(
            "custom Field overrides for the constructor parameters. "
            "See :func:`pydantic.Fields.Field`"
        ),
    )

    annotations: Dict[str, Any] = Field(
        default_factory=dict,
        description="custom annotations for constructor parameters",
    )

    template: Template = Field(
        Template(name="class_.py"),
        description=(
            "The :class:`demessaging.template.Template` that is used "
            "to render the class for the generated API."
        ),
    )

    reporter_args: Dict[str, BaseReport] = Field(
        default_factory=dict,
        description="Arguments that use the dasf-progress-api",
    )


@append_parameter_docs
class BaseMessagingConfig(BaseSettings):
    """Base class for messaging configs."""

    class Config:
        env_prefix = "de_backend_"

    topic: str = Field(
        description=(
            "The topic identifier under which to register at the pulsar."
        )
    )

    header: Union[Json[Dict[str, Any]], Dict[str, Any]] = Field(  # type: ignore
        default_factory=dict, description="Header parameters for the request"
    )

    max_workers: Optional[PositiveInt] = Field(
        default=None,
        description=(
            "(optional) number of concurrent workers for handling requests, "
            "default: number of processors on the machine, multiplied by 5."
        ),
    )

    queue_size: Optional[PositiveInt] = Field(
        default=None,
        description=(
            "(optional) size of the request queue, if MAX_WORKERS is set, "
            "this needs to be at least as big as MAX_WORKERS, "
            "otherwise an AttributeException is raised."
        ),
    )

    max_payload_size: int = Field(
        default=500 * 1024,
        description=(
            "(optional) maximum payload size, must be smaller than pulsars 'webSocketMaxTextFrameSize', "
            "which is configured e.g.via 'pulsar/conf/standalone.conf'."
            "default: 512000 (500kb)."
        ),
    )

    @root_validator
    def validate_queue_size(cls, values):
        """Check that the queue_size is smaller than the max_workers."""
        queue_size = values.get("queue_size")
        max_workers = values.get("max_workers")
        if queue_size is not None and max_workers is not None:
            if queue_size < max_workers:
                raise ValueError(
                    f"queue_size ({queue_size}) needs to be larger than or "
                    f"equal to max_workers ({max_workers})"
                )
        return values

    def get_topic_url(
        self, topic: str, subscription: Optional[str] = None
    ) -> str:
        """Build the URL to connect to a websocket."""
        raise NotImplementedError(
            "this method is supposed to be implemented by subclasses"
        )


@append_parameter_docs
class PulsarConfig(BaseMessagingConfig):
    """A configuration class to connect to the pulsar messaging framework."""

    class Config:
        env_prefix = "de_backend_"

    host: str = Field(
        "localhost", description="The remote host of the pulsar."
    )

    port: str = Field(
        "8080", description="The port of the pulsar at the given :attr:`host`."
    )

    persistent: str = Field("non-persistent")

    tenant: str = Field("public")

    namespace: str = Field("default")

    def get_topic_url(
        self, topic: str, subscription: Optional[str] = None
    ) -> str:
        """Build the URL to connect to a websocket."""
        connection_type = "consumer" if subscription else "producer"
        sub = ("/" + subscription) if subscription else ""
        return (
            f"ws://{self.host}:{self.port}/ws/v2/{connection_type}/"
            f"{self.persistent}/{self.tenant}/{self.namespace}/{topic}{sub}"
        )


@append_parameter_docs
class WebsocketURLConfig(BaseMessagingConfig):
    """A configuration for a websocket."""

    class Config:
        env_prefix = "de_backend_"

    websocket_url: str = Field(
        "", description="The fully qualified URL to the websocket."
    )

    producer_url: Optional[str] = Field(
        None,
        description=(
            "An alternative URL to use for producers. If None, the "
            "`websocket_url` will be used."
        ),
    )

    consumer_url: Optional[str] = Field(
        None,
        description=(
            "An alternative URL to use for consumers. If None, the "
            "`websocket_url` will be used."
        ),
    )

    def get_topic_url(
        self, topic: str, subscription: Optional[str] = None
    ) -> str:
        """Build the URL to connect to a websocket."""
        sub = ("/" + subscription) if subscription else ""
        if subscription:
            uri = self.consumer_url or self.websocket_url
        else:
            uri = self.producer_url or self.websocket_url
        if not uri.endswith("/"):
            uri += "/"
        return uri + topic + sub


@append_parameter_docs
class ModuleConfig(BaseConfig):
    """Configuration class for a backend module."""

    class Config:
        arbitrary_types_allowed = True

    # it should be Type[BackendFunction], Type[BaseModel], but that's
    #  not well supported by pydantic
    if TYPE_CHECKING:
        members: List[
            Union[
                str,
                Callable,
                Type[object],
                Type[
                    BackendFunction  # pylint: disable=used-before-assignment  # noqa: E501
                ],
                Type[BackendClass],  # pylint: disable=used-before-assignment
            ]
        ]

    messaging_config: Union[PulsarConfig, WebsocketURLConfig] = Field(
        description="Configuration on how to connect to the message broker."
    )

    members: List[Union[str, Callable, Type[object], Any]] = Field(  # type: ignore  # noqa: E501
        default_factory=list, description="List of members for this module"
    )

    imports: str = Field(
        "",
        description="Imports that should be added to the generate API module.",
    )

    template: Template = Field(
        Template(name="module.py"),
        description=(
            "The :class:`demessaging.template.Template` that is used "
            "to render the module for the generated API."
        ),
    )

    @property
    def pulsar_config(self) -> Union[PulsarConfig, WebsocketURLConfig]:
        """DEPRECATED! Get the messaging configuration.

        Please use the ``messaging_config`` attribute of this class."""
        warn(
            "The `pulsar_config` property is deprecated. Please use the "
            "`messaging_config` instead",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.messaging_config


def configure(js: Optional[str] = None, **kwargs) -> Callable[[T], T]:
    """Configuration decorator for function or modules.

    Use this function as a decorator for classes or functions in the backend
    module like so::

        >>> @configure(field_params={"a": {"gt": 0}}, returns={"gt": 0})
        ... def sqrt(a: float) -> float:
        ...     import math
        ...
        ...     return math.sqrt(a)


    The available parameters for this function vary depending on what you
    are decorating. If you are decorating a class, your parameters must be
    valid for the :class:`ClassConfig`. If you are decorating a function, your
    parameters must be valid for a :class:`FunctionConfig`.

    Parameters
    ----------
    js: Optional[str]
        A JSON-formatted string that can be used to setup the config.
    ``**kwargs``
        Any keyword argument that can be used to setup the config.

    Notes
    -----
    If you are specifying any ``kwargs``, your first argument (`js`) should
    be ``None``.
    """

    def decorator(obj: T) -> T:
        ConfClass: Union[Type[ClassConfig], Type[FunctionConfig]]
        if inspect.isclass(obj):
            ConfClass = ClassConfig
        else:
            ConfClass = FunctionConfig
        if js and kwargs:
            raise ValueError(
                "You can either specify a JSON string or keyword arguments, "
                "not both!"
            )
        if js:
            config = ConfClass.parse_raw(js)
        else:
            config = ConfClass(**kwargs)
        obj.__pulsar_config__ = config  # type: ignore
        return obj

    return decorator



import unittest

import xarray as xr

from demessaging import NetCDFDecoder


class TestNetCDFDecoder(unittest.TestCase):

    _serialized_path = "serialized.json"
    _original_path = "original.nc"

    def testDecode(self):
        # deserialize test data
        with open(self._serialized_path, "r") as serialized:
            dataset: xr.Dataset = NetCDFDecoder.decode(serialized.read())

        self.assertIsNotNone(dataset, "error decoding serialized netcdf")

        # test decoded version against original
        original = xr.open_dataset(filename_or_obj=self._original_path)

        # check for crs
        self.assertEqual(
            dataset.attrs.get("crs").casefold(),
            original.attrs.get("crs").casefold(),
        )

        # check for data variable
        self.assertIn("daily_rainfall", dataset)

        # check dimensions
        self.assertTupleEqual(
            dataset["daily_rainfall"].dims, original["daily_rainfall"].dims
        )

        # check dimension sizes
        self.assertTupleEqual(
            dataset["daily_rainfall"].shape, original["daily_rainfall"].shape
        )


if __name__ == "__main__":
    unittest.main()



"""Producer of messages submitted for the message broker."""
from __future__ import annotations

import base64
import json
import threading
from datetime import datetime
from itertools import count
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from deprogressapi import BaseReport
from websocket import WebSocket

from demessaging.messaging.connection import WebsocketConnection
from demessaging.PulsarMessageConstants import (
    MessageType,
    PropertyKeys,
    Status,
)

if TYPE_CHECKING:
    from demessaging.config import BaseMessagingConfig


class MessageProducer(WebsocketConnection):
    """Producer class to send requests to a registered backend module (topic)"""

    SOCKET_PING_INTERVAL = 5  # 1min

    def __init__(self, pulsar_config: BaseMessagingConfig):
        super().__init__(pulsar_config)
        self.subscription_name: str = (
            "python-backend-" + datetime.now().isoformat()
        )
        self.context_counter = count()

    def _socket_ping(
        self, sockets: List[WebSocket], timeout, event: threading.Event
    ):
        while not event.wait(timeout):
            for sock in sockets:
                if sock.connected:
                    try:
                        print("ping {}".format(datetime.now()))
                        sock.ping()
                    except Exception as e:
                        print("error in socket ping routine: {}".format(e))
                        break
                else:
                    print(
                        "quitting socket ping loop due to unconnected or missing socket"
                    )
                    break

    def start_ping_loop(self, sockets: List[WebSocket]):
        event = threading.Event()
        thread = threading.Thread(
            target=self._socket_ping,
            args=(sockets, MessageProducer.SOCKET_PING_INTERVAL, event),
        )
        thread.setDaemon(True)
        thread.start()

    async def send_request(
        self, request_msg, topic: Optional[str] = None
    ) -> Any:
        """Sends the given request to the backend module bound to the topic provided in the pulsar configuration.
        In order to increase re-usability the destination topic can be overridden with the optional topic argument.

        :param request_msg: dictionary providing a 'property' dictionary, a payload string, or both
        :param topic: overrides the used topic for this request
        :return: received response from the backend module
        """
        # establish connections
        out_topic = self.pulsar_config.topic
        # topic override if given
        if topic is not None:
            out_topic = topic
        out_socket: WebSocket = self.open_socket(
            topic=out_topic, header=self.pulsar_config.header
        )
        # To be thread safe, we generate the response topic here
        response_topic: str = self.generate_response_topic()
        in_socket: WebSocket = self.open_socket(
            subscription=self.subscription_name,
            topic=response_topic,
            header=self.pulsar_config.header,
        )

        self.start_ping_loop([in_socket, out_socket])

        try:
            # create message context (from counter)
            context = next(self.context_counter)
            request_msg["context"] = context
            if "properties" not in request_msg:
                request_msg["properties"] = {}
            request_msg["properties"][
                PropertyKeys.RESPONSE_TOPIC
            ] = response_topic
            request_msg["properties"][
                PropertyKeys.REQUEST_CONTEXT
            ] = request_msg["context"]
            request_msg["properties"][
                PropertyKeys.MESSAGE_TYPE
            ] = MessageType.REQUEST

            # send message via outgoing connection to request topic
            out_socket.send(json.dumps(request_msg))
            ack = out_socket.recv()
            ack = json.loads(ack)
            if "error" in ack["result"]:
                # error - return
                return {
                    "status": "error",
                    "error": "error sending the request",
                    "msg": ack,
                }

            # wait for response on response topic
            response = in_socket.recv()

            # parse json message
            response = json.loads(response)
            props = response["properties"]

            # acknowledge the response
            in_socket.send(json.dumps({"messageId": response["messageId"]}))

            # mapping from ids to existing reports
            reports: Dict[str, BaseReport] = {}

            while props[PropertyKeys.MESSAGE_TYPE] == MessageType.PROGRESS:
                # we received a progress report - print and ignore
                # decode progress data
                progress_data = base64.b64decode(response["payload"]).decode(
                    "utf-8"
                )

                report = BaseReport.from_payload(progress_data)
                if report.report_id in reports:
                    base_report = reports[report.report_id]
                    for field in report.__fields__:
                        setattr(base_report, field, getattr(report, field))
                else:
                    reports[report.report_id] = base_report = report
                if base_report.status != Status.RUNNING:
                    base_report.complete(base_report.status)
                else:
                    base_report.submit()

                # wait for next response
                response = in_socket.recv()

                # parse json message
                response = json.loads(response)
                props = response["properties"]

            # assert that we received a 'response' message and check for matching context
            if props[PropertyKeys.MESSAGE_TYPE] != MessageType.RESPONSE:
                return {
                    "status": "error",
                    "error": "received message is not a response to the sent request",
                    "msg": response,
                }

            if int(props[PropertyKeys.REQUEST_CONTEXT]) != int(context):
                return {
                    "status": "error",
                    "error": "received message belongs to a different request",
                    "msg": response,
                }

            if "payload" in response:
                status = "success"
                if "status" in props:
                    # we might successfully get a response, but it might contain an error from the backend
                    status = props["status"]

                # decode b64 payload msg
                payload: str = response["payload"]

                try:
                    payload = base64.b64decode(payload).decode("utf-8")
                except Exception as e:
                    status = "error"
                    payload = "error decoding payload: {0}".format(e)

                if status == "error":
                    return {
                        "status": "error",
                        "error": payload,
                        "msg": response,
                    }
                else:
                    return {"status": status, "msg": payload}
            else:
                return {
                    "status": "error",
                    "error": "missing response payload",
                    "msg": response,
                }
        finally:
            # close the sockets
            if out_socket:
                out_socket.close()
            if in_socket:
                in_socket.close()



"""Test backend module

This module exists just for test purposes
"""
import uuid
from typing import List

from conftest import ArbitraryTestModel, ArbitraryType

from demessaging import main

__all__ = [
    "Class",
    "func_arbitrary_model",
    "func_basic",
    "func_arbitrary_type",
]


topic = "test_topic_" + uuid.uuid4().urn[9:]


class Class:
    """A test class."""

    def __init__(self, a: int, b: int = 1):
        """
        Parameters
        ----------
        a : int
            first integer
        b : int, optional
            second integer, by default 1
        """
        self.a = a
        self.b = b

    @staticmethod
    def sum(a: int, b: int) -> int:
        """
        Parameters
        ----------
        a : int
            first integer
        b : int, optional
            second integer, by default 1

        Returns
        -------
        int
            The sum of `a` and `b`
        """
        return a + b

    def _this_should_not_be_included(self, a: int) -> int:
        """This method should not be included because it stats with _

        Parameters
        ----------
        a: int
            Some dummy int
        """
        return a

    def add2a(self, c: int) -> int:
        """Add a number to `a`.

        Parameters
        ----------
        c : int
            The number to add

        Returns
        -------
        int
            a + c
        """
        return self.a + c

    def add2b(self, c: int) -> int:
        """Add a number to `b`

        Parameters
        ----------
        c : int
            The number to add

        Returns
        -------
        int
            b + c
        """
        return self.b + c


def func_basic(a: int) -> List[int]:
    """A test function.

    With some docu.

    Parameters
    ----------
    a: integer
        An integer

    Returns
    -------
    list of int
        A list of integers
    """
    return [a]


def func_arbitrary_type(a: ArbitraryType) -> ArbitraryType:
    """Test function for arbitrary types.

    Parameters
    ----------
    a: ArbitraryType
        Some arbitrary thing

    Returns
    -------
    ArbitraryType
        The same as `a`
    """
    return a


def func_arbitrary_model(a: ArbitraryTestModel) -> ArbitraryTestModel:
    """Test function for arbitrary models.

    Parameters
    ----------
    a: ArbitraryTestModel
        Some arbitrary thing

    Returns
    -------
    ArbitraryTestModel
        The same as `a`
    """
    return a


def _private_func():
    """This should not be included."""
    pass


if __name__ == "__main__":
    main(messaging_config=dict(topic=topic))



"""Test backend module

This module exists just for test purposes
"""
import uuid
from typing import List

from demessaging import config, main

__all__ = ["func_basic"]


topic = "test_topic_" + uuid.uuid4().urn[9:]


@config.registry.register_type
class MyType:
    """Just a test type to see if it appeads in the generated API module."""

    def __init__(self, a: int) -> None:
        self.a = a

    def add2a(self, b: int) -> int:
        return self.a + b


def func_basic(a: int) -> List[int]:
    """A test function.

    With some docu.

    Parameters
    ----------
    a: integer
        An integer

    Returns
    -------
    list of int
        A list of integers
    """
    return [a]


if __name__ == "__main__":
    main(messaging_config=dict(topic=topic))



from warnings import warn

from demessaging.messaging.connection import (  # noqa: F403, F401
    WebsocketConnection,
)

warn(
    "The demessaging.PulsarConnection module has been renamed to "
    "demessaging.messaging.connection and will be removed soon!",
    DeprecationWarning,
)


class PulsarConnection(WebsocketConnection):
    # deprecated

    def __init__(self, *args, **kwargs):
        warn(
            "The `demessaging.PulsarConnection.PulsarConnection` class has "
            "been replaced by the "
            "`demessaging.messaging.connection.WebsocketConnection` class "
            "and will be removed soon!",
            DeprecationWarning,
        )
        super().__init__(*args, **kwargs)



"""DASF Methods to connect to the Message Broker."""

from .connection import WebsocketConnection  # noqa: F401
from .consumer import MessageConsumer  # noqa: F401
from .producer import MessageProducer  # noqa: F401



"""Transform a python function into a corresponding pydantic model.

The :class:`BackendFunction` model in this module generates subclasses based
upon a python class (similarly as the
:class:`~demessaging.backend.class_.BackendClass` does it for classes).
"""
from __future__ import annotations

import inspect
import warnings
from functools import partial
from textwrap import dedent
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Dict,
    Optional,
    Type,
    cast,
)

import docstring_parser
from pydantic import Field  # pylint: disable=no-name-in-module
from pydantic import BaseModel, create_model
from pydantic.json import (  # pylint: disable=no-name-in-module
    custom_pydantic_encoder,
)

import demessaging.backend.utils as utils
from demessaging.config import FunctionConfig, append_parameter_docs


def get_return_model(
    docstring: docstring_parser.Docstring, config: BackendFunctionConfig
) -> Type[BaseModel]:
    """Generate field for the return property.

    Parameters
    ----------
    docstring : docstring_parser.Docstring
        The parser that analyzed the docstring

    Returns
    -------
    Any
        The pydantic field
    """
    return_description = ""
    ret_count: int = 0
    for arg in docstring.meta:
        if (
            isinstance(arg, docstring_parser.DocstringReturns)
            and arg.description
        ):
            return_description += "\n- " + arg.description
            ret_count += 1
    return_description = return_description.strip()
    if ret_count == 1:
        return_description = return_description[2:]

    field_kws: Dict[str, Any] = {"default": None}

    if return_description.strip():
        field_kws["description"] = return_description

    field_kws.update(config.returns)

    ret_field = Field(**field_kws)  # type: ignore

    sig = config.signature

    Model: Type[BaseModel]

    create_kws: Dict[str, Any] = {}
    if config.registry.json_encoders:

        class Config:
            json_encoders = config.registry.json_encoders

        create_kws["__config__"] = Config

    if sig and sig.return_annotation is not sig.empty:
        if sig.return_annotation is None:
            create_kws["__root__"] = (Any, ret_field)
        else:
            create_kws["__root__"] = (sig.return_annotation, ret_field)
    else:
        warnings.warn(
            f"Missing return signature for {config.function.__name__}!",
            RuntimeWarning,
        )
        create_kws["__root__"] = (Any, ret_field)
    Model = create_model(
        config.class_name,
        **create_kws,  # type: ignore
    )

    return Model


@append_parameter_docs
class BackendFunctionConfig(FunctionConfig):
    """Configuration class for a backend module function."""

    function: Any = Field(description="The function to call.")

    class_name: str = Field(description="Name of the model class")

    def update_from_function(self) -> None:
        """Update the config from the corresponding function."""
        func = self.function
        if not self.name:
            self.name = func.__name__ or ""
        if not self.doc:
            self.doc = dedent(inspect.getdoc(func) or "")
        if not self.signature:
            self.signature = inspect.signature(func)


@append_parameter_docs
class BackendFunction(BaseModel):
    """A base class for a function model.

    Don't use this model, rather use :meth:`create_model` method to
    generate new models.
    """

    class Config:
        validate_assignment = True

    backend_config: ClassVar[BackendFunctionConfig]

    return_model: ClassVar[Type[BaseModel]]

    if TYPE_CHECKING:
        # added properties for subclasses generated by create_model
        func_name: str

    def __call__(self) -> BaseModel:  # type: ignore
        kws = utils.get_kws(self.backend_config.signature, self)

        for key in self.backend_config.reporter_args:
            kws[key] = getattr(self, key)

        ret = self.backend_config.function(**kws)

        return self.return_model.parse_obj(ret)

    @classmethod
    def create_model(
        cls,
        func: Callable,
        config: Optional[FunctionConfig] = None,
        class_name=None,
        **kwargs,
    ) -> Type[BackendFunction]:
        """Create a new pydantic Model from a function.

        Parameters
        ----------
        func: callable
            A function or method
        config: FunctionConfig, optional
            The configuration to use. If given, this overrides the
            ``__pulsar_config__`` of the given `func`
        class_name: str, optional
            The name for the generated subclass of :class:`pydantic.BaseModel`.
            If not given, the name of `func` is used
        ``**kwargs``
            Any other parameter for the :func:`pydantic.create_model` function

        Returns
        -------
        Subclass of BackendFunction
            The newly generated class that represents this function.
        """
        sig = inspect.signature(func)
        docstring = docstring_parser.parse(func.__doc__)  # type: ignore

        if config is None:
            config = getattr(func, "__pulsar_config__", FunctionConfig())
        config = cast(FunctionConfig, config)

        name = cast(str, func.__name__)
        if not class_name:
            class_name = utils.snake_to_camel("Func", name)

        config = BackendFunctionConfig(
            function=func,
            class_name=class_name,
            **config.copy(deep=True).dict(),
        )

        config.update_from_function()

        fields = utils.get_fields(name, sig, docstring, config)

        desc = utils.get_desc(docstring)

        kwargs.update(fields)

        Model: Type[BackendFunction] = create_model(  # type: ignore
            class_name,
            __validators__=config.validators,
            __base__=cls,
            __module__=func.__module__,
            **kwargs,  # type: ignore
        )

        Model.return_model = get_return_model(docstring, config)

        Model.backend_config = config

        if config.registry.json_encoders:
            # it would be better, to set this via __config__ in create_model,
            # but this is not possible if we use `__base__`
            Model.__config__.json_encoders = config.registry.json_encoders
            Model.__json_encoder__ = partial(
                custom_pydantic_encoder, config.registry.json_encoders
            )

        if desc:
            Model.__doc__ = desc
        else:
            Model.__doc__ = ""

        return Model


BackendFunctionConfig.update_forward_refs()



import pathlib
from dataclasses import field
from typing import Any, Dict

import jinja2
from pydantic.dataclasses import dataclass  # pylint: disable=no-name-in-module

TEMPLATE_FOLDER = pathlib.Path(__file__).parent / "templates"


@dataclass
class Template:
    """A convenience wrapper to render a jinja2 template."""

    name: str
    folder: pathlib.Path = TEMPLATE_FOLDER
    suffix: str = ".jinja2"
    context: Dict[str, Any] = field(default_factory=dict)

    @property
    def path(self) -> pathlib.Path:
        return self.folder / (self.name + self.suffix)

    @property
    def renderer(self) -> Any:
        loader = jinja2.FileSystemLoader(str(self.folder), followlinks=True)
        env = jinja2.Environment(loader=loader)
        return env.get_template(self.name + self.suffix)

    def render(self, **context) -> str:
        context.update(self.context)
        context["set"] = set
        return self.renderer.render(**context)



"""Backend module to transform a python module into a pydantic model.

This module defines the main model in the demessaging framework. It takes a
list of members, or a module, and creates a new Model that can be used to
generate code, connect to the pulsar, and more. See :class:`BackendModule` for
details.
"""
from __future__ import annotations

import atexit
import base64
import datetime as dt
import inspect
import io
from functools import partial
from importlib import import_module
from typing import (
    IO,
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Dict,
    List,
    Optional,
    Type,
    Union,
    cast,
)

import docstring_parser
from deprogressapi import BaseReport
from pydantic import Field  # pylint: disable=no-name-in-module
from pydantic import BaseModel, ValidationError, create_model
from pydantic.json import (  # pylint: disable=no-name-in-module
    custom_pydantic_encoder,
)

from demessaging.backend import utils
from demessaging.backend.class_ import BackendClass
from demessaging.backend.function import BackendFunction
from demessaging.config import ModuleConfig, append_parameter_docs
from demessaging.messaging.consumer import MessageConsumer
from demessaging.PulsarMessageConstants import PropertyKeys, Status


@append_parameter_docs
class BackendModuleConfig(ModuleConfig):
    """Configuration class for a backend module."""

    # it should be Union[Type[BackendFunction], Type[BackendClass]], but
    # this is not supported by pydantic

    if TYPE_CHECKING:
        models: List[Union[Type[BackendFunction], Type[BackendClass]]]

    models: List[Any] = Field(  # type: ignore
        default_factory=list,
        description=(
            "a list of function or class models for the members of the "
            "backend module"
        ),
    )

    module: Any = Field(
        description="The imported backend module (or none, if there is none)"
    )

    class_name: str = Field(description="Name of the model class")


ModuleMember = Union[
    Type[BackendFunction], Type[BackendClass], Callable, str, Type[object]
]


@append_parameter_docs
class BackendModule(BaseModel):
    """A base class for a backend module.

    Do not directly instantiate from this class, rather use the
    :meth:`create_model` method.
    """

    backend_config: ClassVar[BackendModuleConfig]
    pulsar: ClassVar[MessageConsumer]

    # type that is implemented by subclasses
    __root__: Union[BackendFunction, BackendClass]

    if TYPE_CHECKING:
        # added properties for subclasses generated by create_model
        member: Union[BackendClass, BackendFunction]

    def __call__(self) -> BaseModel:
        """Call the selected member of this backend module."""
        return self.__root__()  # type: ignore

    @classmethod
    def create_model(
        cls,
        module_name: Optional[str] = None,
        members: Optional[List[ModuleMember]] = None,
        config: Optional[ModuleConfig] = None,
        class_name: Optional[str] = None,
        **config_kws,
    ) -> Type[BackendModule]:
        """Generate a module for a backend module.

        Parameters
        ----------
        module_name: str
            The name of the module to import. If none is given, the `members`
            must be specified
        members: list of members
            The list of members that shall be added to this module. It can be
            a list of

            - :class:`~demessaging.backend.function.BackendFunction` classes (
              generated with
              :meth:`~demessaging.backend.function.BackendFunction.create_model`)
            - :class:`~demessaging.backend.class_.BackendClass` classes (
              generated with
              :meth:`~demessaging.backend.class_.BackendClass.create_model`)
            - functions (that will then be transformed using
              :meth:`~demessaging.backend.function.BackendFunction.create_model`)
            - classes (that will then be transformed using
              :meth:`~demessaging.backend.class_.BackendClass.create_model`)
            - strings, in which case they point to the member of the given
              `module_name`
        config: ModuleConfig, optional
            The configuration for the module. If this is not given, you must
            provide ``config_kws`` or define a ``backend_config`` variable
            within the module corresponding to `module_name`
        class_name: str, optional
            The name for the generated subclass of :class:`pydantic.BaseModel`.
            If not given, the name of `Class` is used
        ``**config_kws``
            An alternative way to specify the configuration for the backend
            module.

        Returns
        -------
        Subclass of BackendFunction
            The newly generated class that represents this module.
        """
        if module_name is not None:
            module: Any = import_module(module_name)
        else:
            module = None

        if members is None and module is None:
            raise ValueError("Either members or module need to be provided!")

        if config and config_kws:
            raise ValueError("Either config or config_kws can be used!")
        if config_kws:
            config = ModuleConfig(**config_kws)
        elif module is not None and hasattr(module, "backend_config"):
            config = module.backend_config

        config = cast(ModuleConfig, config)

        # this should not be camelized
        class_name = class_name or module_name or config.messaging_config.topic

        assert config is not None
        config = BackendModuleConfig(
            module=module, class_name=class_name, **config.copy().dict()
        )

        if not members:
            members = list(config.members)
        if not members:
            assert module is not None
            if hasattr(module, "__all__"):
                members = list(module.__all__)
            else:
                functions = inspect.getmembers(
                    module, predicate=inspect.isfunction
                )
                classes = inspect.getmembers(
                    module, predicate=inspect.isfunction
                )
                members = [t[1] for t in functions if not t[0].startswith("_")]
                members += [t[1] for t in classes if not t[0].startswith("_")]

        # finally check if we have any members
        if not members:
            raise ValueError(
                f"Found no members for the given module {module_name}!"
            )
        models: List[Union[Type[BackendFunction], Type[BackendClass]]] = []

        for i, member in enumerate(list(members)):
            member_obj: ModuleMember
            member_model: Union[Type[BackendFunction], Type[BackendClass]]
            if isinstance(member, str):
                member = getattr(module, member)
            if inspect.isclass(member) and issubclass(
                member, (BackendFunction, BackendClass)  # type: ignore
            ):
                member = cast(
                    Union[Type[BackendFunction], Type[BackendClass]], member
                )
                member_model = member
                member_obj = (
                    member.backend_config.Class
                    if issubclass(member, BackendClass)
                    else member.backend_config.function
                )
            elif inspect.isclass(member):
                member_model = BackendClass.create_model(member)
                member_obj = member
            elif callable(member):
                member_model = BackendFunction.create_model(member)
                member_obj = member
            else:
                raise ValueError(
                    f"Cannot transform {member} to a member model!"
                )
            members[i] = member_obj
            models.append(member_model)

        config.members = members
        config.models = models

        if not config.doc and module:
            docstring = docstring_parser.parse(module.__doc__)
            config.doc = utils.get_desc(docstring)

        member_types = models[0]
        for model in models[1:]:
            member_types = Union[member_types, model]  # type: ignore

        kws = {"__module__": module_name} if module_name else {}

        Model: Type[BackendModule] = create_model(  # type: ignore
            class_name,
            __base__=cls,
            __root__=(member_types, Field(description="The member to call.")),
            **kws,  # type: ignore
        )

        Model.__config__.title = config.messaging_config.topic  # type: ignore

        if config.registry.json_encoders:
            # it would be better, to set this via __config__ in create_model,
            # but this is not possible if we use `__base__`
            Model.__config__.json_encoders = config.registry.json_encoders
            Model.__json_encoder__ = partial(
                custom_pydantic_encoder, config.registry.json_encoders
            )

        Model.backend_config = config

        if module is not None:
            config.imports += "\n" + utils.get_module_imports(module)

        Model.__doc__ = config.doc

        return Model

    @classmethod
    def test_connect(cls):
        """Connect to the message pulsar."""
        cls.pulsar = pulsar = MessageConsumer(
            pulsar_config=cls.backend_config.messaging_config,
            handle_request=cls.handle_message,
            module_info=cls.schema(),
        )
        atexit.register(pulsar.disconnect)

        pulsar.connect()

    @classmethod
    def listen(cls):
        """Connect to the message pulsar."""
        cls.pulsar = pulsar = MessageConsumer(
            pulsar_config=cls.backend_config.messaging_config,
            handle_request=cls.handle_message,
            module_info=cls.schema(),
        )
        atexit.register(pulsar.disconnect)

        pulsar.connect()
        pulsar.wait_for_request()

    @classmethod
    async def _send_request(cls, request: Dict[str, Any]) -> Any:
        """Send a request to the backend module.

        Parameters
        ----------
        request: dict
            A request to the backend module.
        """
        from demessaging.messaging.producer import MessageProducer

        producer = MessageProducer(cls.backend_config.messaging_config)

        response = await producer.send_request(request)
        status = response[PropertyKeys.STATUS]
        if status == Status.SUCCESS:
            print("request successful")
            return response["msg"]
        elif status == Status.ERROR:
            print("request failed")
            print(response["msg"])
            raise ValueError(response["error"])

    @classmethod
    def send_request(
        cls: Type[BackendModule],
        request: Union[BackendModule, IO, Dict[str, Any]],
    ) -> BaseModel:
        """Test a request to the backend.

        Parameters
        ----------
        request: dict or file-like object
            A request to the backend module.
        """
        if isinstance(request, io.IOBase):
            model = cls.parse_raw("\n".join(request.readlines()))
        elif hasattr(request, "__root__"):
            request = cast(BackendModule, request)
            model = cls.parse_obj(request.__root__)
        else:
            model = cls.parse_obj(request)
        payload = base64.b64encode(model.json().encode("utf-8")).decode(
            "utf-8"
        )
        request = {
            "properties": {},
            "payload": payload,
        }
        result = utils.run_async(cls._send_request, request)

        return model.__root__.return_model.parse_raw(result)

    def compute(self) -> BaseModel:
        """Send this request to the backend module and compute the result.

        This method updates the model inplace.
        """
        response = self.send_request(self)
        return response

    @classmethod
    def shell(cls):
        """Start a shell with the module defined."""
        from IPython import start_ipython

        start_ipython(argv=[], user_ns=dict(Model=cls))

    @classmethod
    def generate(
        cls,
        line_length: int = 79,
        use_formatters: bool = True,
        use_autoflake: bool = True,
        use_black: bool = True,
        use_isort: bool = True,
    ) -> str:
        """Generate the code for the frontend module."""
        import autoflake
        import black
        import isort

        code = cls.backend_config.render()

        if use_formatters:

            if use_isort:
                code = isort.code(code, float_to_top=True, profile="black")
            if use_black:
                code = black.format_str(
                    code, mode=black.Mode(line_length=line_length)
                )

            # remove unused imports
            if use_autoflake:
                code = autoflake.fix_code(code, remove_all_unused_imports=True)

            if use_isort:
                code = isort.code(code, float_to_top=True, profile="black")

        if cls.backend_config.module:
            # remove __main__, etc.
            name = cls.backend_config.module.__name__
            code = code.replace(name + ".", "")

        return code.strip() + "\n"

    @classmethod
    def handle_message(cls, request_msg):
        print("[{}] processing request".format(dt.datetime.now()))

        payload = base64.b64decode(request_msg["payload"]).decode("utf-8")

        try:
            model = cls.parse_raw(payload)
        except ValidationError as e:
            cls.pulsar.send_error(
                request=request_msg,
                error_message="error validating request: {0}".format(e),
            )
        except Exception as e:
            cls.pulsar.send_error(
                request=request_msg,
                error_message="error processing request: {0}".format(e),
            )
        else:
            try:
                reporter_args = model.__root__.backend_config.reporter_args
                for key, reporter in reporter_args.items():
                    member_reporter = getattr(model.__root__, key)
                    if member_reporter and isinstance(
                        member_reporter, BaseReport
                    ):
                        member_reporter._pulsar = cls.pulsar
                        member_reporter._request = request_msg
                result = model()
            except Exception as e:
                cls.pulsar.send_error(
                    request=request_msg,
                    error_message="error executing request: {0}".format(e),
                )
            else:
                cls.pulsar.send_response(
                    request=request_msg,
                    response_properties={PropertyKeys.STATUS: Status.SUCCESS},
                    response_payload=result.json(),
                )


ModuleConfig.update_forward_refs()



"""Enums within the DASF Messaging Framework."""
from enum import Enum
from typing import Type
from warnings import warn


def generate_enum_doc(class_: Type[Enum]) -> Type[Enum]:
    """Utility function to generate the docstring for an enum."""
    docstring = "\n\nThe following values are valid:\n\n"

    params = [
        f'- **{name}** (``"{prop.value}"``)'
        for name, prop in class_.__members__.items()
    ]

    class_.__doc__ += docstring + "\n".join(params)  # type: ignore
    return class_


@generate_enum_doc
class Status(str, Enum):
    """Status flag of a request."""

    SUCCESS = "success"
    ERROR = "error"
    RUNNING = "running"


@generate_enum_doc
class PropertyKeys(str, Enum):
    """Property keys for a message to the message broker."""

    REQUEST_CONTEXT = "requestContext"
    RESPONSE_TOPIC = "response_topic"
    SOURCE_TOPIC = "source_topic"
    REQUEST_MESSAGEID = "requestMessageId"
    MESSAGE_TYPE = "messageType"
    FRAGMENT = "fragment"
    NUM_FRAGMENTS = "num_fragments"
    STATUS = "status"


@generate_enum_doc
class MessageType(str, Enum):
    """Supported message types."""

    PING = "ping"
    PONG = "pong"
    REQUEST = "request"
    RESPONSE = "response"
    LOG = "log"
    INFO = "info"
    PROGRESS = "progress"


class _DeprecatedPulsarConfigKeys:
    """DEPRECATED PulsarConfigKeys"""

    def __getattr__(self, attr):
        attrs = dict(
            HOST="host",
            PORT="port",
            PERSISTENT="persistent",
            TENANT="tenant",
            NAMESPACE="namespace",
            TOPIC="topic",
            MAX_WORKERS="max_workers",
            QUEUE_SIZE="queue_size",
            MAX_PAYLOAD_SIZE="max_payload_size",
        )
        try:
            ret = attrs[attr]
        except KeyError:
            raise AttributeError(attr)
        else:
            warn(
                "The PulsarConfigKeys class has been deprecated and will be "
                "removed soon!",
                DeprecationWarning,
                stacklevel=2,
            )
            return ret


PulsarConfigKeys = _DeprecatedPulsarConfigKeys()



"""Test module for the :mod:`demessaging.backend` module."""
import importlib
import inspect
import pathlib
from textwrap import dedent
from typing import Any, Callable, Dict

import pytest
from conftest import ArbitraryType
from pydantic import ValidationError  # pylint: disable=no-name-in-module
from pytest_lazyfixture import lazy_fixture as lf

from demessaging import backend

try:
    import xarray
except ImportError:
    xarray = None  # type: ignore
else:
    try:
        import netCDF4
    except ImportError:
        netCDF4 = None  # type: ignore


class TestFunctionModel:
    """Test functions for creating a function model."""

    def test_missing_doc(self, func_missing_doc: Callable) -> None:
        """Test parsing a function with missing docstrings."""
        Model = backend.BackendFunction.create_model(func_missing_doc)
        schema = Model.schema()
        assert "description" not in schema
        assert "description" not in schema["properties"]["a"]

    def test_missing_sig(self, func_missing_sig: Callable) -> None:
        """Test parsing a function with missing signatures."""
        with pytest.warns(RuntimeWarning, match="Missing signature for a"):
            Model = backend.BackendFunction.create_model(func_missing_sig)
        schema = Model.schema()
        assert "type" not in schema["properties"]["a"]

    @pytest.mark.skip("Not yet adapted")
    def test_missing_return_sig(self, func_missing_sig: Callable) -> None:
        """Test parsing a function with missing signatures."""
        with pytest.warns(RuntimeWarning, match="Missing return signature"):
            Model = backend.BackendFunction.create_model(func_missing_sig)
        schema = Model.schema()
        assert "func_returns" in schema["properties"]
        assert "description" in schema["properties"]["func_returns"]
        desc = schema["properties"]["func_returns"]["description"]
        assert desc == "An integer"

    def test_arbitraty_types(self, func_arbitrary_types: Callable) -> None:
        Model = backend.BackendFunction.create_model(func_arbitrary_types)
        schema = Model.schema()

        # test the properties
        assert "type" not in schema["properties"]["a"]
        assert "custom_type" in schema["properties"]["a"]
        assert schema["properties"]["a"]["custom_type"] == (
            "conftest.ArbitraryType"
        )

        # test the return statement
        # assert "func_returns" in schema["properties"]
        # assert "type" not in schema["properties"]["func_returns"]
        # assert "custom_type" in schema["properties"]["func_returns"]
        # assert schema["properties"]["func_returns"]["custom_type"] == (
        #     "conftest.ArbitraryType"
        # )

    def test_arbitraty_model(self, func_arbitrary_model: Callable) -> None:
        """Test input and output of arbitraty pydantic models."""
        Model = backend.BackendFunction.create_model(func_arbitrary_model)
        schema = Model.schema()

        assert "ArbitraryTestModel" in schema["definitions"]
        assert "$ref" in schema["properties"]["a"]
        assert schema["properties"]["a"]["$ref"].endswith("ArbitraryTestModel")

        assert "ArbitraryTestModel" in schema["definitions"]
        # assert "$ref" in schema["properties"]["func_returns"]
        # assert schema["properties"]["func_returns"]["$ref"].endswith(
        #     "ArbitraryTestModel"
        # )

    @pytest.mark.parametrize(
        "func,obj,xfail",
        [
            (lf("func_sig"), lf("valid_obj"), False),
            (lf("func_sig"), lf("invalid_obj"), True),
            (lf("func_missing_doc"), lf("valid_obj"), False),
            (lf("func_missing_doc"), lf("invalid_obj"), True),
            (lf("func_missing_sig"), lf("valid_obj"), False),
            (lf("func_missing_sig"), lf("invalid_obj"), False),
            (lf("func_arbitrary_types"), lf("valid_obj"), False),
            (lf("func_arbitrary_types"), lf("invalid_obj"), True),
            (lf("func_arbitrary_model"), lf("valid_arbitrary_model"), False),
            (lf("func_arbitrary_model"), lf("invalid_obj"), True),
        ],
    )
    @pytest.mark.filterwarnings("ignore: Missing signature")
    @pytest.mark.filterwarnings("ignore: Missing return signature")
    def test_function_request(
        self, func: Callable, obj: Dict[str, Any], xfail: bool
    ) -> None:
        """Test parsing a function to the model."""
        Model = backend.BackendFunction.create_model(func)
        obj["func_name"] = func.__name__

        if xfail:
            with pytest.raises(ValidationError):
                Model.parse_obj(obj)
        else:
            Model.parse_obj(obj)

    def test_invalid_func_name(
        self, func_sig: Callable, valid_obj: Dict[str, Any]
    ) -> None:
        """Test if the function cannot be parsed if a wrong name is given."""
        Model = backend.BackendFunction.create_model(func_sig)
        valid_obj["func_name"] = func_sig.__name__ + "123"

        with pytest.raises(ValidationError):
            Model.parse_obj(valid_obj)

    def test_missing_func_name(
        self, func_sig: Callable, valid_obj: Dict[str, Any]
    ) -> None:
        """Test if the function cannot be parsed if a wrong name is given."""
        Model = backend.BackendFunction.create_model(func_sig)

        with pytest.raises(ValidationError):
            Model.parse_obj(valid_obj)

    def test_call_function(
        self, func_arbitrary_types: Callable, valid_obj: Dict[str, Any]
    ) -> None:
        """Test calling a function with conversion to arbitrary type."""
        Model = backend.BackendFunction.create_model(func_arbitrary_types)
        valid_obj["func_name"] = func_arbitrary_types.__name__
        model = Model.parse_obj(valid_obj)
        ret = model()
        assert isinstance(ret.__root__, ArbitraryType)  # type: ignore
        assert ret.__root__.a == 1  # type: ignore

    def test_call_function_2(
        self, func_sig: Callable, valid_obj: Dict[str, Any]
    ) -> None:
        """Test calling a function with conversion to list."""
        Model = backend.BackendFunction.create_model(func_sig)
        valid_obj["func_name"] = func_sig.__name__
        model = Model.parse_obj(valid_obj)
        ret = model()
        assert ret.__root__ == [1]  # type: ignore

    def test_call_invalid_function(
        self, func_invalid_sig: Callable, valid_obj: Dict[str, Any]
    ) -> None:
        """Test calling a function with invalid signature."""
        Model = backend.BackendFunction.create_model(func_invalid_sig)
        valid_obj["func_name"] = func_invalid_sig.__name__
        model = Model.parse_obj(valid_obj)
        with pytest.raises(ValidationError):
            model()

    def test_render(
        self,
        func_sig: Callable,
        tmp_module: pathlib.Path,
        random_mod_name: str,
    ) -> None:
        """Test the rendering of a function."""
        Model = backend.BackendFunction.create_model(func_sig)
        code = Model.backend_config.render()

        tmp_module.write_text("from __future__ import annotations\n" + code)

        mod = importlib.import_module(random_mod_name)

        assert hasattr(mod, func_sig.__name__)
        func = getattr(mod, func_sig.__name__)

        ref_doc = dedent(inspect.getdoc(func_sig)).strip()  # type: ignore
        func_doc = dedent(inspect.getdoc(func)).strip()  # type: ignore

        assert func_doc == ref_doc



import re

from pygments.lexer import bygroups, default, include
from pygments.lexers.javascript import TypeScriptLexer
from pygments.token import Name, Operator, Punctuation, String, Text

# Use same tokens as `TypeScriptLexer`, but with tags and attributes support
TOKENS = TypeScriptLexer.tokens

TOKENS.update(
    {
        "vue": [
            (r"(<)([\w-]+)", bygroups(Punctuation, Name.Tag), "tag"),
            (
                r"(<)(/)([\w-]+)(>)",
                bygroups(Punctuation, Punctuation, Name.Tag, Punctuation),
            ),
        ],
        "tag": [
            (r"\s+", Text),
            (
                r"([@:]?[\w-]+\s*)(=)(\s*)",
                bygroups(Name.Attribute, Operator, Text),
                "attr",
            ),
            (r"[{}]+", Punctuation),
            (r"[\w\.-]+", Name.Attribute),
            (
                r"(/?)(\s*)(>)",
                bygroups(Punctuation, Text, Punctuation),
                "#pop",
            ),
        ],
        "attr": [
            ("{", Punctuation, "expression"),
            ('".*?"', String, "#pop"),
            ("'.*?'", String, "#pop"),
            default("#pop"),
        ],
        "expression": [
            ("{", Punctuation, "#push"),
            ("}", Punctuation, "#pop"),
            include("root"),
        ],
    }
)
TOKENS["root"].insert(0, include("vue"))


class VueTypescriptLexer(TypeScriptLexer):
    name = "vue"
    aliases = ["vue", "vuejs"]
    filenames = ["*.vue"]
    mimetypes = ["text/x-vue", "application/x-vue"]

    flags = re.MULTILINE | re.DOTALL | re.UNICODE

    tokens = TOKENS


def setup(app):
    app.add_lexer("vue", VueTypescriptLexer)



"""Test backend module

This module exists just for test purposes
"""
import os
import uuid
from typing import List

from pydantic import BaseModel

from demessaging import config, main

__all__ = ["func_basic", "TestClass"]


topic = os.getenv("TEST_TOPIC", "test_topic_" + uuid.uuid4().urn[9:])


@config.registry.register_type
class TestBaseModelType(BaseModel):
    """Just a test type to see if it appeads in the generated API module."""

    data: int


def func_basic(a: List[TestBaseModelType]) -> int:
    """A test function.

    With some docu.

    Parameters
    ----------
    a: integer
        An integer

    Returns
    -------
    list of int
        A list of integers
    """
    return a[0].data


class TestClass:
    """A class to test custom basemodel types."""

    def __init__(self, a: List[TestBaseModelType]):
        """
        Parameters
        ----------
        a: List[TestBaseModelType]
            One data object
        """
        self.a = a

    def test_func(self, b: TestBaseModelType) -> int:
        """A test method

        Parameters
        ----------
        b : TestBaseModelType
            Some more data

        Returns
        -------
        int
            the sum of both
        """
        adata = self.a[0].data
        return adata + b.data


if __name__ == "__main__":
    main(messaging_config=dict(topic=topic))



"""Consumer for messages submitted via the message broker."""
from __future__ import annotations

import asyncio
import base64
import concurrent.futures
import json
import sys
import textwrap
import threading
from datetime import datetime
from sys import stderr
from time import sleep
from typing import TYPE_CHECKING, Dict, List, Optional

import websocket

from demessaging.messaging.connection import WebsocketConnection
from demessaging.PulsarMessageConstants import MessageType, PropertyKeys

if TYPE_CHECKING:
    from demessaging.config import BaseMessagingConfig

# patch the asyncio loop if we are on windows
# see https://github.com/tornadoweb/tornado/issues/2751
if sys.platform == "win32":
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())


class MessageConsumer(WebsocketConnection):
    """Consumer for messages submitted via the message broker."""

    INITIAL_TIMEOUT_SLEEP = 10  # seconds
    MAX_CONNECTION_ATTEMPTS = 20
    PULSAR_PING_INTERVAL = 60  # 1min

    def __init__(
        self,
        pulsar_config: BaseMessagingConfig,
        handle_request,
        handle_response=None,
        module_info: Optional[dict] = None,
    ):
        super().__init__(pulsar_config)
        pulsar_config = self.pulsar_config
        self.handle_request = handle_request
        self.handle_response = handle_response
        self.module_info = module_info

        # init event loop
        self.loop = asyncio.get_event_loop()
        self.request_semaphore: Optional[threading.BoundedSemaphore] = None
        if pulsar_config.queue_size is not None:
            self.request_semaphore = threading.BoundedSemaphore(
                value=pulsar_config.queue_size
            )
        self.pool = concurrent.futures.ThreadPoolExecutor(
            max_workers=pulsar_config.max_workers
        )
        self.send_lock = threading.Lock()

        self.max_payload_size = pulsar_config.max_payload_size
        self.connection_attempts = 0
        self.reconnectTimeout = MessageConsumer.INITIAL_TIMEOUT_SLEEP
        self.subscription = None
        self.producers: Dict[str, websocket.WebSocket] = {}

    def connect(self):
        # disconnect if already connected
        if self.subscription:
            self.disconnect()

        # prepare timestamp string as part of subscription name
        timestr = datetime.now().isoformat()[:19]
        subscription_name = "backend-module-" + timestr

        # create consumer socket subscription
        while (
            not self.connection_attempts
            > MessageConsumer.MAX_CONNECTION_ATTEMPTS
        ):
            try:
                self.connection_attempts += 1
                print("connection attempt {}".format(self.connection_attempts))
                self.subscription = self.open_socket(
                    subscription=subscription_name,
                    header=self.pulsar_config.header,
                )

                if self.subscription:
                    # successful connection - reset timeout and attempt count
                    self.connection_attempts = 0
                    self.reconnectTimeout = (
                        MessageConsumer.INITIAL_TIMEOUT_SLEEP
                    )
                    self.start_ping_loop()
                    return
            except Exception as e:
                print(
                    f"unable to establish connection (reason: {e}), next attempt in {self.reconnectTimeout} sec."
                )
                # unable to connect - wait and try again
                sleep(self.reconnectTimeout)
                # increase the timeout for the next attempt
                self.reconnectTimeout *= 2

    def disconnect(self):
        # close consumer
        if self.subscription:
            try:
                self.subscription.close()
            except Exception as e:
                print(
                    "warn: error while closing subscription socket: {0}".format(
                        e
                    )
                )
            finally:
                self.subscription = None

        # close producers
        for producer in self.producers.values():
            try:
                producer.close()
            except Exception as e:
                print(
                    "warn: error while closing producer socket: {0}".format(e)
                )

        self.producers.clear()

    def _pulsar_ping(self, timeout, event: threading.Event):
        while not event.wait(timeout):
            if self.subscription and self.subscription.connected:
                try:
                    # print('ping {}'.format(datetime.now()))
                    self.subscription.ping()
                except Exception as e:
                    print("error in pulsar ping routine: {}".format(e))
                    self.reconnect()
                    break
            else:
                print(
                    "quitting pulsar ping loop due to unconnected or missing subscription"
                )
                break

    def start_ping_loop(self):
        event = threading.Event()
        thread = threading.Thread(
            target=self._pulsar_ping,
            args=(MessageConsumer.PULSAR_PING_INTERVAL, event),
        )
        thread.setDaemon(True)
        thread.start()

    def wait_for_request(self):
        # register request event handler
        print("waiting for incoming request")
        self.loop.add_reader(self.subscription, self.receive_request)

        try:
            self.loop.run_forever()
        except KeyboardInterrupt:
            print("keyboard interrupt received")
        finally:
            print("shutting down event loop and disconnecting sockets")
            if self.loop.is_running():
                self.loop.stop()
            # self.loop.close()
            self.disconnect()

    def reconnect(self):
        print("reconnect, attempt {0}".format(self.connection_attempts))

        # do we exceed the maximum number of connection attempts
        if self.connection_attempts > self.MAX_CONNECTION_ATTEMPTS:
            print(
                "exceeding maximum connection attempts: {0}".format(
                    self.connection_attempts
                )
            )
            self.loop.stop()
            self.disconnect()
        elif self.subscription:
            # there already is a subscription - remove it from the loop
            # in case the subscription is not connected anymore the reader has been already removed
            if self.subscription.connected:
                # only try to remove the subscription reader if still connected
                try:
                    self.loop.remove_reader(self.subscription)
                except ValueError:
                    print(
                        "ignoring exception in remove reader during reconnect"
                    )

            # disconnect and (re-)connect
            self.connect()

            # add the subscription to the running loop
            self.loop.add_reader(self.subscription, self.receive_request)

    def receive_request(self):
        if not self.subscription or not self.subscription.connected:
            print("subscription connection lost")
            # the socket is not connected anymore
            self.reconnect()
            return

        try:
            # receive json message
            msg = self.subscription.recv()

            # handle empty message
            if msg is None:
                print("empty message received")
                return

            # parse json message
            msg = json.loads(msg)

            # validate message
            # verify that we got a response_topic
            if MessageConsumer.is_valid_request(msg):
                # acknowledge request
                # FIXME: when do we actually acknowledge a message? right after receiving it or after processing it?
                self.acknowledge(msg)

                # handle according to message type
                msg_type = MessageConsumer.extract_message_type(msg)
                if msg_type == MessageType.PING:
                    # simply reply with pong
                    self.send_pong(msg)
                elif msg_type == MessageType.PONG:
                    # handle pong message
                    self.handle_pong(msg)
                elif msg_type == MessageType.INFO:
                    # handle info message
                    self.handle_info(msg)
                elif msg_type == MessageType.REQUEST:
                    # handle request message later via event loop
                    # self.loop.call_soon(self.handle_request, msg)
                    if self.request_semaphore is None:
                        # no bounded queue - directly pass the request to the executor
                        self.loop.run_in_executor(
                            self.pool, self.handle_request, msg
                        )
                    else:
                        # bounded queue size - check available space
                        if self.request_semaphore.acquire(blocking=False):
                            # there is still space in the queue - pass to executor queue
                            self.loop.run_in_executor(
                                self.pool, self.handle_request_via_queue, msg
                            )
                        else:
                            # queue is full - reject request
                            self.send_error(
                                request=msg,
                                error_message="request rejected due to queue overflow, try again later",
                            )

                    # todo: do we need to address any exceptions here?? e.g. via future.add_done_callback()
                    # Process(target=self.handle_request, args=(msg,)).start()
                elif msg_type == MessageType.RESPONSE:
                    if self.handle_response:
                        # handle response message later via event loop
                        self.loop.call_soon(self.handle_response, msg)
                        # Process(target=self.handle_response, args=(msg,)).start()
                    else:
                        print(
                            "ignoring response message due to missing handle_response function"
                        )
                else:
                    print("received unsupported message type: " + msg_type)
            else:
                print(
                    "message with unsupported structure received - ignoring it"
                )
        except SystemError:
            print("receive interrupted")
            self.loop.stop()
            return
        except Exception as e:
            print(
                "error in receive message loop {0} - trying to reconnect".format(
                    e
                )
            )
            # reconnect
            self.reconnect()

    def handle_request_via_queue(self, msg):
        # here the semaphore has already been acquired
        try:
            self.handle_request(msg)
        finally:
            # no matter what - release the semaphore
            self.request_semaphore.release()

    def acknowledge(self, msg):
        self.subscription.send(json.dumps({"messageId": msg["messageId"]}))

    def send_error(self, request, error_message):
        self.send_response(
            request=request,
            response_payload=error_message,
            response_properties={"status": "error"},
        )

    def send_response(
        self,
        request,
        response_payload=None,
        msg_type=MessageType.RESPONSE,
        response_properties=None,
    ):
        with self.send_lock:
            # validate original request
            if MessageConsumer.is_valid_request(request):
                # the request is valid - create a producer for the given response topic
                response_topic = MessageConsumer.extract_response_topic(
                    request
                )

                if response_properties is None:
                    response_properties = {}

                # prepare response
                response_properties[
                    PropertyKeys.REQUEST_CONTEXT
                ] = MessageConsumer.extract_context(request)
                response_properties[PropertyKeys.MESSAGE_TYPE] = msg_type
                response_properties[
                    PropertyKeys.SOURCE_TOPIC
                ] = self.pulsar_config.topic

                if response_topic in self.producers:
                    # we already have a producer for this
                    producer = self.producers[response_topic]
                else:
                    # no producer yet, create one
                    producer = self.open_socket(
                        topic=response_topic, header=self.pulsar_config.header
                    )
                    self.producers[response_topic] = producer

                # send the response
                msg = {"properties": response_properties, "payload": ""}
                if response_payload:
                    if isinstance(response_payload, str):
                        response_payload = response_payload.encode("utf-8")
                    elif isinstance(response_payload, dict):
                        response_payload = json.dumps(response_payload).encode(
                            "utf-8"
                        )

                    if len(response_payload) > self.max_payload_size:
                        print("sending fragmented message")
                        # the payload exceeds the maximum size - send fragmented message
                        payload_fragments: List[str] = textwrap.wrap(
                            text=base64.b64encode(response_payload).decode(
                                "utf-8"
                            ),
                            width=self.max_payload_size,
                        )

                        num_fragments = len(payload_fragments)
                        for i in range(num_fragments):
                            fragment_props = response_properties.copy()
                            fragment_props[PropertyKeys.FRAGMENT] = i
                            fragment_props[
                                PropertyKeys.NUM_FRAGMENTS
                            ] = num_fragments

                            fragmented_msg = {
                                "properties": fragment_props,
                                "payload": payload_fragments[i],
                            }
                            producer.send(json.dumps(fragmented_msg))

                            # receive acknowledgment
                            ack = json.loads(producer.recv())
                            if ack["result"] != "ok":
                                print(
                                    "Failed to send message: {}".format(ack),
                                    file=stderr,
                                )

                            # sleep 100ms
                            sleep(0.05)

                        print("sent {} fragments".format(num_fragments))
                        return

                    # send unfragmented message
                    msg["payload"] = base64.b64encode(response_payload).decode(
                        "utf-8"
                    )

                producer.send(json.dumps(msg))

                # receive acknowledgment
                # fixme: is this thread proof???
                #  what happens if the same producer send multiple messages at the same time?
                #  how can we assign the acknowledgment to the sent message?
                ack = json.loads(producer.recv())
                if ack["result"] != "ok":
                    print(
                        "Failed to send message: {}".format(ack), file=stderr
                    )

    def send_pong(self, request):
        self.send_response(request=request, msg_type=MessageType.PONG)

    def handle_pong(self, request):
        print("pong received {0}", request)

    def handle_info(self, info_request):
        # check for available module info
        if self.module_info is None:
            # no module info provided
            print("missing info")
            self.send_response(
                info_request,
                response_properties={
                    "info": "This module does not provide capability information."
                },
            )
            return

        print("sending info response...")
        self.send_response(
            request=info_request,
            response_properties={"info": json.dumps(self.module_info)},
        )

    @staticmethod
    def extract_response_topic(msg):
        if PropertyKeys.RESPONSE_TOPIC in msg["properties"]:
            return msg["properties"][PropertyKeys.RESPONSE_TOPIC]
        else:
            return None

    @staticmethod
    def extract_context(msg):
        if PropertyKeys.REQUEST_CONTEXT in msg["properties"]:
            return str(msg["properties"][PropertyKeys.REQUEST_CONTEXT])
        else:
            return None

    @staticmethod
    def extract_message_type(msg):
        if PropertyKeys.MESSAGE_TYPE in msg["properties"]:
            return msg["properties"][PropertyKeys.MESSAGE_TYPE]
        else:
            return None

    @staticmethod
    def is_valid_value(value):
        return value is not None and isinstance(value, str) and len(value) > 0

    @staticmethod
    def is_valid_request(request_message):
        # for now the request is valid if we find a valid response topic, context and message type
        return (
            MessageConsumer.is_valid_value(
                MessageConsumer.extract_response_topic(request_message)
            )
            and MessageConsumer.is_valid_value(
                MessageConsumer.extract_context(request_message)
            )
            and MessageConsumer.is_valid_value(
                MessageConsumer.extract_message_type(request_message)
            )
        )



"""
Example backend module for the de-messaging framework.

Start this module by typing::

    python ExampleMessageConsumer.py -t your-custom-topic connect

in the terminal, or::

    python ExampleMessageConsumer.py --help

for further options.
"""
import datetime
from typing import Any, Callable, Dict, List, Union

import deprogressapi.tree
from deprogressapi import ProgressReport
from pydantic import BaseModel

import demessaging.types.xarray
from demessaging import BackendModule as _BackendModule
from demessaging import configure, main
from demessaging.config import ModuleConfig

NoneType = type(None)


__all__ = ["hello_world", "HelloWorld", "compute_sum", "report_progress_demo"]


class GreetResponse(BaseModel):

    message: str
    greetings: List[str]
    greeting_time: datetime.datetime


@configure(
    """
{
    "field_params": {
        "repeat": {
            "ge": 0
        }
    }
}
"""
)
def hello_world(
    message: str, repeat: int, greet_message: str
) -> GreetResponse:
    """
    Greet the hello world module.

    Parameters
    ----------
    message: str
        Message that will be part of the greet response
    repeat: int
        Number of repetitions of the greet message
    greet_message: str
        Message string that will be mirror back to the requesting node

    Returns
    -------
    GreetResponse
        Hello world greet response object
    """
    request = {
        "func_name": "hello_world",
        "message": message,
        "repeat": repeat,
        "greet_message": greet_message,
    }

    model = BackendModule.parse_obj(request)
    response = model.compute()

    return response.__root__  # type: ignore


class HelloWorld:
    """
    Greet the world from a class.

    Classes can define the methods that shall be used for the backend, and they
    define a constructor.
    """

    def __init__(self, message: str):
        """
        Parameters
        ----------
        message: str
            The hello message
        """
        self._request_base: Dict[str, Any] = {
            "class_name": "HelloWorld",
            "message": message,
        }

    @configure(
        """
    {
        "field_params": {
            "repeat": {
                "ge": 0
            }
        }
    }
    """
    )
    def repeat_message(self, repeat: int) -> List[str]:
        """
        Repeat the `message` multiple times.

        Parameters
        ----------
        repeat: int
            The number of times, `message` shall be repeated.
        """
        request = self._request_base.copy()
        request["function"] = {"func_name": "repeat_message", "repeat": repeat}

        model = BackendModule.parse_obj(request)
        response = model.compute()

        return response.__root__  # type: ignore


def compute_sum(
    da: demessaging.types.xarray.DataArray,
    reporter: deprogressapi.tree.ProgressReport = ProgressReport(
        report_id="root", step_message="Sum computation", steps=0, children=[]
    ),
) -> demessaging.types.xarray.DataArray:
    """
    Compute the sum over a data array.

    Parameters
    ----------
    da : DataArray
        The input data array

    Returns
    -------
    DataArray
        The sum of the data array
    """
    request = {"func_name": "compute_sum", "da": da, "reporter": reporter}

    model = BackendModule.parse_obj(request)
    response = model.compute()

    return response.__root__  # type: ignore


def report_progress_demo(
    reporter: Union[
        deprogressapi.tree.ProgressReport, NoneType
    ] = ProgressReport(
        report_id="root", step_message="Demo report", steps=0, children=[]
    )
) -> None:
    """
    Demo function for a progress report with the dasf-progress-api.
    """
    request = {"func_name": "report_progress_demo", "reporter": reporter}

    model = BackendModule.parse_obj(request)
    response = model.compute()

    return response.__root__  # type: ignore


backend_config = ModuleConfig.parse_raw(
    """
{
    "messaging_config": {
        "topic": "test-topic",
        "header": {},
        "max_workers": 4,
        "queue_size": 4,
        "max_payload_size": 512000,
        "websocket_url": "ws://127.0.0.1:8000/ws/",
        "producer_url": "",
        "consumer_url": ""
}
}
"""
)

_creator: Callable
if __name__ == "__main__":
    _creator = main
else:
    _creator = _BackendModule.create_model

BackendModule = _creator(
    __name__,
    config=backend_config,
    class_name="BackendModule",
    members=[hello_world, HelloWorld, compute_sum, report_progress_demo],
)



from warnings import warn

from demessaging.messaging.constants import *  # noqa: F403, F401

warn(
    "The demessaging.PulsarMessageConstants module has been renamed to "
    "demessaging.messaging.constants and will be removed soon!",
    DeprecationWarning,
)



"""Custom types module.

This module contains several convenience types to be used with pydantic.
"""


class TypeMixin:
    @classmethod
    def __get_validators__(cls):
        yield cls.validate  # pylint: disable=no-member

    @classmethod
    def __modify_schema__(cls, field_schema):
        type_ = cls.__module__ + "." + cls.__name__
        field_schema["custom_type"] = type_


class TypeMixinSlots:
    """Alternative way for TypeMixin, because xarray objects need __slots__."""

    __slots__ = ()

    @classmethod
    def __get_validators__(cls):
        yield cls.validate  # pylint: disable=no-member

    @classmethod
    def __modify_schema__(cls, field_schema):
        type_ = cls.__module__ + "." + cls.__name__
        field_schema["custom_type"] = type_
